from warnings import warn

import numpy as np
import pandas as pd

from ..weights.raster import _da2wsp
from ._utils import (
    _sparse_to_arrays,
)


def _raster_contiguity(
    da,
    criterion="queen",
    z_value=None,
    coords_labels=None,
    k=1,
    include_nodata=False,
    n_jobs=1,
):
    """
    Create an input for Graph from xarray.DataArray.

    Parameters
    ----------
    da : xarray.DataArray
        Input 2D or 3D DataArray with shape=(z, y, x)
    criterion : {"rook", "queen"}
        Type of contiguity. Default is queen.
    z_value : int/string/float
        Select the z_value of 3D DataArray with multiple layers.
    coords_labels : dictionary
        Pass dimension labels for coordinates and layers if they do not
        belong to default dimensions, which are (band/time, y/lat, x/lon)
        e.g. coords_labels = {"y_label": "latitude",
        "x_label": "longitude", "z_label": "year"}
        Default is {} empty dictionary.
    k : int
        Order of contiguity, this will select all neighbors upto kth order.
        Default is 1.
    include_nodata : boolean
        If True, missing values will be assumed as non-missing when
        selecting higher_order neighbors, Default is False
    n_jobs : int
        Number of cores to be used in the sparse weight construction. If -1,
        all available cores are used. Default is 1.

    Returns
    -------
    (head, tail, weight, index_names)


    """
    try:
        import numba  # noqa: F401

        use_numba = True
        include_nodata = False
    except (ModuleNotFoundError, ImportError):
        warn(
            "numba cannot be imported, parallel processing "
            "and include_nodata functionality will be disabled. "
            "falling back to slower method",
            stacklevel=2,
        )
        use_numba = False

    if coords_labels is None:
        coords_labels = {}

    if use_numba:
        (weight, (head, tail)), ser, _ = _da2wsp(
            da=da,
            criterion=criterion,
            z_value=z_value,
            coords_labels=coords_labels,
            k=k,
            include_nodata=include_nodata,
            n_jobs=n_jobs,
            use_numba=use_numba,
        )
        order = np.lexsort((tail, head))
        head = head[order]
        tail = tail[order]
        weight = weight[order]

        head = ser.index.to_numpy()[head]
        tail = ser.index.to_numpy()[tail]
    else:
        sw, ser = _da2wsp(
            da=da,
            criterion=criterion,
            z_value=z_value,
            coords_labels=coords_labels,
            k=k,
            include_nodata=include_nodata,
            n_jobs=n_jobs,
            use_numba=use_numba,
        )
        head, tail, weight = _sparse_to_arrays(sw, ser.index.to_numpy())

    return (
        head,
        tail,
        weight,
        ser.index,
    )


def _generate_da(g, y):
    """Creates xarray.DataArray object from passed data aligned with the Graph.

    Parameters
    ----------
    g : Graph
        Graph, ideally generated using _raster_contiguity builder to ensure it
        contains _xarray_index_names attribute.
    y : array_like
        flat array that shall be reshaped into a DataArray with dimensionality
        conforming to Graph

    Returns
    -------
    xarray.DataArray
        instance of xarray.DataArray that can be aligned with the DataArray from which
        Graph was built
    """
    if hasattr(g, "_xarray_index_names"):
        names = g._xarray_index_names
    else:
        warn(
            UserWarning,
            "Graph does not store xarray index names."
            "The output may not align with the original DataArray.",
            stacklevel=3,
        )
        names = None
    return pd.Series(
        y,
        index=pd.MultiIndex.from_tuples(g.unique_ids, names=names),
    ).to_xarray()



# ruff: noqa: N802, N815, SIM115

import scipy.io as sio

from ...weights import W
from ...weights.util import full, full2W
from .. import fileio

__author__ = "Myunghwa Hwang <mhwang4@gmail.com>"
__all__ = ["MatIO"]


class MatIO(fileio.FileIO):
    """Opens, reads, and writes weights file objects in MATLAB Level 4-5 MAT format.
    ``.mat`` files are used in Dr. LeSage's MATLAB Econometrics library. The ``.mat``
    file format can handle both full and sparse matrices, and it allows for a matrix
    dimension greater than 256. In PySAL, row and column headers of a MATLAB array
    are ignored.

    PySAL uses `scipy.io <https://docs.scipy.org/doc/scipy/reference/tutorial/io.html>`_.
    Thus, it is subject to all limits of ``scipy.io.loadmat`` and ``scipy.io.savemat``.

    Notes
    -----

    If a given weights object contains too many observations to write it out as
    a full matrix, PySAL writes out the object as a sparse matrix.

    References
    ----------

    `MathWorks <http://www.mathworks.com/help/pdf_doc/matlab/matfile_format.pdf>`_
    (2011) "MATLAB 7 MAT-File Format."

    """

    FORMATS = ["mat"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        self._varName = "Unknown"
        fileio.FileIO.__init__(self, *args, **kwargs)
        self.file = open(self.dataPath, self.mode + "b")

    def _set_varName(self, val):
        if issubclass(type(val), str):
            self._varName = val

    def _get_varName(self) -> str:
        return self._varName

    varName = property(fget=_get_varName, fset=_set_varName)

    def read(self, n=-1):  # noqa: ARG002
        """

        Parameters
        ----------
        n : int
            Read at most ``n`` objects. Default is ``-1``.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        """

        self._complain_ifclosed(self.closed)

        w = self._read()

        return w

    def seek(self, pos):
        if pos == 0:
            self.file.seek(0)
            self.pos = 0

    def _read(self):
        """Reads MATLAB ``.mat`` file.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        Examples
        --------

        Type ``dir(w)`` at the interpreter to see what methods are supported.
        Open a MATLAB ``.mat`` file and read it into a PySAL weights object.

        >>> import libpysal
        >>> w = libpysal.io.open(libpysal.examples.get_path(
        ...     'spat-sym-us.mat'), 'r'
        ... ).read()

        Get the number of observations from the header.

        >>> w.n
        46

        Get the mean number of neighbors.

        >>> w.mean_neighbors
        4.086956521739131

        Get neighbor distances for a single observation.

        >>> w[1] == dict({25: 1, 3: 1, 28: 1, 39: 1})
        True

        """

        if self.pos > 0:
            raise StopIteration

        mat = sio.loadmat(self.file)
        mat_keys = [k for k in mat if not k.startswith("_")]
        full_w = mat[mat_keys[0]]

        self.pos += 1

        w = full2W(full_w)

        return w

    def write(self, obj):
        """Write a weights object to the opened MATLAB ``.mat`` file.

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        TypeError
            Raised when the input ``obj`` is not a PySAL `W`.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(
        ...     libpysal.examples.get_path('spat-sym-us.mat'), 'r'
        ... )
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='.mat')

        Reassign to a new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w')

        Write the weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created ``.mat`` file.

        >>> wnew =  libpysal.io.open(fname, 'r').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)

        if issubclass(type(obj), W):
            try:
                w = full(obj)[0]
            except ValueError:
                w = obj.sparse
            sio.savemat(self.file, {"WEIGHT": w})
            self.pos += 1
        else:
            raise TypeError(f"Expected a PySAL weights object, got: {type(obj)}.")

    def close(self):
        self.file.close()
        fileio.FileIO.close(self)



import numpy as np
import pytest

from libpysal import graph
from libpysal.graph._spatial_lag import _lag_spatial
from libpysal.weights import lat2W


class TestLag:
    def setup_method(self):
        self.neighbors = {
            "a": ["b"],
            "b": ["c", "a"],
            "c": ["b"],
            "d": [],
        }
        self.weights = {"a": [1.0], "b": [1.0, 1.0], "c": [1.0], "d": []}
        self.g = graph.Graph.from_dicts(self.neighbors, self.weights)
        self.y = np.array([0, 1, 2, 3])
        self.yc = np.array([*"ababcbcbc"])
        w = lat2W(3, 3)
        w.transform = "r"
        self.gc = graph.Graph.from_W(w)

    def test_lag_spatial(self):
        yl = _lag_spatial(self.g, self.y)
        np.testing.assert_array_almost_equal(yl, [1.0, 2.0, 1.0, 0])
        g = graph.Graph.from_W(lat2W(3, 3))
        y = np.arange(9)
        yl = _lag_spatial(g, y)
        ylc = np.array([4.0, 6.0, 6.0, 10.0, 16.0, 14.0, 10.0, 18.0, 12.0])
        np.testing.assert_array_almost_equal(yl, ylc)
        g_row = g.transform("r")
        yl = _lag_spatial(g_row, y)
        ylc = np.array([2.0, 2.0, 3.0, 3.33333333, 4.0, 4.66666667, 5.0, 6.0, 6.0])
        np.testing.assert_array_almost_equal(yl, ylc)

    def test_lag_spatial_categorical(self):
        yl = _lag_spatial(self.gc, self.yc)
        ylc = np.array(["b", "a", "b", "c", "b", "c", "b", "c", "b"], dtype=object)
        np.testing.assert_array_equal(yl, ylc)
        self.yc[3] = "a"  # create ties
        np.random.seed(12345)
        yl = _lag_spatial(self.gc, self.yc, categorical=True, ties="random")
        ylc = np.array(["a", "a", "b", "c", "b", "c", "b", "c", "b"], dtype=object)
        yl1 = _lag_spatial(self.gc, self.yc, categorical=True, ties="random")
        yls = _lag_spatial(self.gc, self.yc, categorical=True, ties="tryself")
        np.testing.assert_array_equal(yl, ylc)
        yl1c = np.array(["b", "a", "b", "c", "b", "c", "b", "c", "b"], dtype=object)
        np.testing.assert_array_equal(yl1, yl1c)
        ylsc = np.array(["a", "a", "b", "c", "b", "c", "a", "c", "b"], dtype=object)
        np.testing.assert_array_equal(yls, ylsc)
        # self-weight
        neighbors = self.gc.neighbors
        neighbors[0] = (0, 3, 1)  # add self neighbor for observation 0
        gc = graph.Graph.from_dicts(neighbors)
        self.yc[3] = "b"
        yls = _lag_spatial(gc, self.yc, categorical=True, ties="tryself")
        assert yls[0] in ["b", "a"]
        self.yc[3] = "a"
        yls = _lag_spatial(gc, self.yc, categorical=True, ties="tryself")
        assert yls[0] == "a"

    def test_ties_raise(self):
        with pytest.raises(ValueError, match="There are 2 ties that must be broken"):
            self.yc[3] = "a"  # create ties
            _lag_spatial(self.gc, self.yc, categorical=True)

    def test_categorical_custom_index(self):
        expected = np.array(["bar", "foo", "bar", "foo"])
        np.testing.assert_array_equal(
            expected, self.g.lag(["foo", "bar", "foo", "foo"])
        )



"""pyrtree Unittest."""

from ..rtree import Rect, RTree


class TestPyrtree:
    def setup_method(self):
        k = 10
        w = 20
        objects = {}
        id_ = 0
        for i in range(k):
            mn_y = i * w
            mx_y = mn_y + w
            for j in range(k):
                mn_x = j * w
                mx_x = mn_x + w
                objects[id_] = Rect(mn_x, mn_y, mx_x, mx_y)
                id_ += 1
        self.objects = objects

    def test_rtree(self):
        t = RTree()
        for object_ in self.objects:
            t.insert(object_, self.objects[object_])
        assert len(self.objects) == 100

        qr = Rect(5, 5, 25, 25)

        # find objects with mbrs intersecting with qr
        res = [r.leaf_obj() for r in t.query_rect(qr) if r.is_leaf()]
        assert len(res) == 4
        res.sort()
        assert res == [0, 1, 10, 11]

        # vertices are shared by all coincident rectangles
        res = [r.leaf_obj() for r in t.query_point((20.0, 20.0)) if r.is_leaf()]
        assert len(res) == 4

        res = [r.leaf_obj() for r in t.query_point((21, 20)) if r.is_leaf()]
        assert len(res) == 2

        # single internal point
        res = [r.leaf_obj() for r in t.query_point((21, 21)) if r.is_leaf()]
        assert len(res) == 1

        # single external point
        res = [r.leaf_obj() for r in t.query_point((-12, 21)) if r.is_leaf()]
        assert len(res) == 0

        qr = Rect(5, 6, 65, 7)

        res = [r.leaf_obj() for r in t.query_rect(qr) if r.is_leaf()]
        assert len(res) == 4



from ...weights import W
from . import gwt

__author__ = "Myunghwa Hwang <mhwang4@gmail.com>"
__all__ = ["DatIO"]


class DatIO(gwt.GwtIO):
    """Opens, reads, and writes file objects in ``.dat`` format. Spatial weights
    objects in ``.dat`` format are used in Dr. LeSage's MatLab Econ library.
    This ``.dat`` format is a simple text file with a ``.DAT`` or ``.dat``
    extension. Without a header line, it includes three data columns
    for origin ID, destination ID, and weight values as follows:

    ```
    [Line 1]    2    1    0.25
    [Line 2]    5    1    0.50
    ```

    Origin/destination IDs in this file format are simply record
    numbers starting with 1. IDs are not necessarily integers.
    Data values for all columns should be numeric.

    """

    FORMATS = ["dat"]
    MODES = ["r", "w"]

    def _read(self):
        """Reads in a ``.dat`` file as a PySAL `W` object.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        Examples
        --------

        Type ``dir(w)`` at the interpreter to see what methods are supported.
        Open ``.dat`` file and read it into a PySAL weights object,

        >>> import libpysal
        >>> w = libpysal.io.open(libpysal.examples.get_path('wmat.dat'), 'r').read()

        Get the number of observations from the header.

        >>> w.n
        49

        Get the mean number of neighbors.

        >>> w.mean_neighbors
        4.73469387755102

        Get neighbor distances for a single observation.

        >>> w[1] == dict({2.0: 0.3333, 5.0: 0.3333, 6.0: 0.3333})
        True

        """

        if self.pos > 0:
            raise StopIteration

        id_type = float
        weights, neighbors = self._readlines(id_type)

        self.pos += 1

        w = W(neighbors, weights)

        return w

    def write(self, obj):
        """Write a weights object to the opened ``.dat`` file.

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(libpysal.examples.get_path('wmat.dat'), 'r')
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='.dat')

        Reassign to a new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w')

        Write the weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created ``.dat`` file.

        >>> wnew =  libpysal.io.open(fname, 'r').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up the temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)

        if issubclass(type(obj), W):
            self._writelines(obj)
        else:
            raise TypeError(f"Expected a PySAL weights object, got: {type(obj)}.")



import pytest

from ....cg.shapes import Chain, Point, Polygon
from ..wkt import WKTParser


class TesttestWKTParser:
    def setup_method(self):
        # Create some Well-Known Text objects
        self.wktPOINT = "POINT(6 10)"
        self.wktLINESTRING = "LINESTRING(3 4,10 50,20 25)"
        self.wktPOLYGON = "POLYGON((1 1,5 1,5 5,1 5,1 1),(2 2, 3 2, 3 3, 2 3,2 2))"
        self.unsupported = [
            "MULTIPOINT(3.5 5.6,4.8 10.5)",
            "MULTILINESTRING((3 4,10 50,20 25),(-5 -8,-10 -8,-15 -4))",
            (
                "MULTIPOLYGON(((1 1,5 1,5 5,1 5,1 1),"
                "(2 2, 3 2, 3 3, 2 3,2 2)),((3 3,6 2,6 4,3 3)))"
            ),
            "GEOMETRYCOLLECTION(POINT(4 6),LINESTRING(4 6,7 10))",
            "POINT ZM (1 1 5 60)",
            "POINT M (1 1 80)",
        ]
        self.empty = ["POINT EMPTY", "MULTIPOLYGON EMPTY"]
        self.parser = WKTParser()

    def test_point(self):
        pt = self.parser(self.wktPOINT)
        assert issubclass(type(pt), Point)
        assert pt[:] == (6.0, 10.0)

    def test_line_string(self):
        line = self.parser(self.wktLINESTRING)
        assert issubclass(type(line), Chain)
        parts = [[pt[:] for pt in part] for part in line.parts]
        assert parts == [[(3.0, 4.0), (10.0, 50.0), (20.0, 25.0)]]
        assert line.len == 73.455384532199886

    def test_polygon(self):
        poly = self.parser(self.wktPOLYGON)
        assert issubclass(type(poly), Polygon)
        parts = [[pt[:] for pt in part] for part in poly.parts]
        assert parts == [
            [(1.0, 1.0), (1.0, 5.0), (5.0, 5.0), (5.0, 1.0), (1.0, 1.0)],
            [(2.0, 2.0), (2.0, 3.0), (3.0, 3.0), (3.0, 2.0), (2.0, 2.0)],
        ]
        assert poly.centroid == (2.9705882352941178, 2.9705882352941178)
        assert poly.area == 17.0

    def test_from_wkt(self):
        for wkt in self.unsupported:
            pytest.raises(NotImplementedError, self.parser.fromWKT, wkt)
        for wkt in self.empty:
            assert self.parser.fromWKT(wkt) is None
        assert self.parser.__call__ == self.parser.fromWKT



import os
import string
import tempfile

import geodatasets
import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
from packaging.version import Version
from scipy import __version__ as scipy_version
from scipy import sparse

from libpysal import graph, weights


@pytest.mark.network
class TestBase:
    def setup_method(self):
        self.neighbor_dict_int = {0: 1, 1: 2, 2: 5, 3: 4, 4: 5, 5: 8, 6: 7, 7: 8, 8: 7}
        self.weight_dict_int_binary = {
            0: 1,
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: 1,
            6: 1,
            7: 1,
            8: 1,
        }
        self.index_int = pd.Index(
            [0, 1, 2, 3, 4, 5, 6, 7, 8],
            dtype="int64",
            name="focal",
        )
        self.neighbor_dict_str = {
            string.ascii_letters[k]: string.ascii_letters[v]
            for k, v in self.neighbor_dict_int.items()
        }
        self.weight_dict_str_binary = {
            string.ascii_letters[k]: v for k, v in self.weight_dict_int_binary.items()
        }
        self.index_str = pd.Index(
            [string.ascii_letters[k] for k in self.index_int],
            dtype="object",
            name="focal",
        )
        self.adjacency_int_binary = pd.Series(
            self.weight_dict_int_binary.values(),
            name="weight",
            index=pd.MultiIndex.from_arrays(
                [self.index_int, self.neighbor_dict_int.values()],
                names=["focal", "neighbor"],
            ),
        )
        self.adjacency_str_binary = pd.Series(
            self.weight_dict_str_binary.values(),
            name="weight",
            index=pd.MultiIndex.from_arrays(
                [self.index_str, self.neighbor_dict_str.values()],
                names=["focal", "neighbor"],
            ),
        )

        # one isolate, one self-link
        self.W_dict_int = {
            0: {0: 1, 3: 0.5, 1: 0.5},
            1: {0: 0.3, 4: 0.3, 2: 0.3},
            2: {1: 0.5, 5: 0.5},
            3: {0: 0.3, 6: 0.3, 4: 0.3},
            4: {1: 0.25, 3: 0.25, 7: 0.25, 5: 0.25},
            5: {2: 0.3, 4: 0.3, 8: 0.3},
            6: {3: 0.5, 7: 0.5},
            7: {4: 0.3, 6: 0.3, 8: 0.3},
            8: {5: 0.5, 7: 0.5},
            9: {},
        }
        self.W_dict_str = {
            string.ascii_letters[k]: {
                string.ascii_letters[k_]: v_ for k_, v_ in v.items()
            }
            for k, v in self.W_dict_int.items()
        }
        self.g_int = graph.Graph.from_weights_dict(self.W_dict_int)
        self.g_str = graph.Graph.from_weights_dict(self.W_dict_str)
        rng = np.random.default_rng(seed=0)
        self.letters = np.asarray(list(string.ascii_letters[:26]))
        rng.shuffle(self.letters)
        self.W_dict_str_unordered = {
            self.letters[k]: {self.letters[k_]: v_ for k_, v_ in v.items()}
            for k, v in self.W_dict_int.items()
        }
        self.g_str_unodered = graph.Graph.from_weights_dict(self.W_dict_str_unordered)

        self.nybb = gpd.read_file(geodatasets.get_path("nybb")).set_index("BoroName")
        self.guerry = gpd.read_file(geodatasets.get_path("geoda guerry"))

    def test_init(self):
        g = graph.Graph(self.adjacency_int_binary)
        assert isinstance(g, graph.Graph)
        assert hasattr(g, "_adjacency")
        assert g._adjacency.shape == (9,)
        pd.testing.assert_series_equal(g._adjacency, self.adjacency_int_binary)
        assert hasattr(g, "transformation")
        assert g.transformation == "O"

        g = graph.Graph(self.adjacency_str_binary)
        assert isinstance(g, graph.Graph)
        assert hasattr(g, "_adjacency")
        assert g._adjacency.shape == (9,)
        pd.testing.assert_series_equal(g._adjacency, self.adjacency_str_binary)
        assert hasattr(g, "transformation")
        assert g.transformation == "O"

        with pytest.raises(TypeError, match="The adjacency table needs to be"):
            graph.Graph(self.adjacency_int_binary.values)

        with pytest.raises(ValueError, match="The index of the adjacency table"):
            adj = self.adjacency_int_binary.copy()
            adj.index.names = ["foo", "bar"]
            graph.Graph(adj)

        with pytest.raises(
            ValueError, match="The adjacency needs to be named 'weight'"
        ):
            graph.Graph(self.adjacency_int_binary.rename("foo"))

        with pytest.raises(ValueError, match="The 'weight' needs"):
            graph.Graph(self.adjacency_int_binary.astype(str))

        with pytest.raises(
            ValueError, match="The adjacency table cannot contain missing"
        ):
            adj = self.adjacency_int_binary.copy()
            adj.iloc[0] = np.nan
            graph.Graph(adj)

        with pytest.raises(ValueError, match="'transformation' needs to be"):
            graph.Graph(self.adjacency_int_binary, transformation="foo")

    def test___repr__(self):
        expected = (
            "<Graph of 10 nodes and 25 nonzero edges indexed by\n"
            " [0, 1, 2, 3, 4, ...]>"
        )
        assert repr(self.g_int) == expected

        expected = (
            "<Graph of 10 nodes and 25 nonzero edges indexed by\n"
            " ['a', 'b', 'c', 'd', 'e', ...]>"
        )
        assert repr(self.g_str) == expected

        nybb = graph.Graph.build_contiguity(self.nybb)
        expected = (
            "<Graph of 5 nodes and 10 nonzero edges indexed by\n"
            " ['Staten Island', 'Queens', 'Brooklyn', 'Manhattan', 'Bronx']>"
        )
        assert repr(nybb) == expected

        h3 = {
            "821f87fffffffff": ("821fb7fffffffff", "821f97fffffffff"),
            "821fb7fffffffff": (
                "821f87fffffffff",
                "821f97fffffffff",
                "82186ffffffffff",
                "821867fffffffff",
            ),
            "821f97fffffffff": (
                "821f87fffffffff",
                "821fb7fffffffff",
                "823967fffffffff",
                "82396ffffffffff",
                "82186ffffffffff",
            ),
            "823967fffffffff": (
                "821f97fffffffff",
                "82396ffffffffff",
                "82186ffffffffff",
            ),
            "82396ffffffffff": ("821f97fffffffff", "823967fffffffff"),
            "82186ffffffffff": (
                "821fb7fffffffff",
                "821f97fffffffff",
                "823967fffffffff",
                "821867fffffffff",
            ),
            "821867fffffffff": ("821fb7fffffffff", "82186ffffffffff"),
        }
        h3_g = graph.Graph.from_dicts(h3)
        expected = (
            "<Graph of 7 nodes and 22 nonzero edges indexed by\n"
            " ['821f87fffffffff', '821fb7fffffffff', '821f97fffffffff',"
            " '823967fffffff...]>"
        )
        assert repr(h3_g) == expected

    def test_copy(self):
        g_copy = self.g_str.copy()
        assert g_copy == self.g_str
        g_copy._adjacency.iloc[0] = 2
        assert g_copy != self.g_str

    def test_adjacency(self):
        g = graph.Graph(self.adjacency_int_binary)
        adjacency = g.adjacency
        pd.testing.assert_series_equal(adjacency, self.adjacency_int_binary)

        # ensure copy
        adjacency.iloc[0] = 100
        pd.testing.assert_series_equal(g._adjacency, self.adjacency_int_binary)

    def test_w_roundtrip(self):
        w = self.g_int.to_W()
        pd.testing.assert_series_equal(
            self.g_int._adjacency.sort_index(),
            w.to_adjlist(drop_islands=False)
            .set_index(["focal", "neighbor"])["weight"]
            .sort_index(),
            check_index_type=False,
            check_dtype=False,
        )
        g_roundtripped = graph.Graph.from_W(w)
        assert self.g_int == g_roundtripped
        assert isinstance(w.id_order, list)

        w = self.g_str.to_W()
        pd.testing.assert_series_equal(
            self.g_str._adjacency.sort_index(),
            w.to_adjlist(drop_islands=False)
            .set_index(["focal", "neighbor"])["weight"]
            .sort_index(),
            check_index_type=False,
            check_dtype=False,
        )
        g_roundtripped = graph.Graph.from_W(w)
        assert self.g_str == g_roundtripped

        w = weights.lat2W(3, 3)
        g = graph.Graph.from_W(w)
        pd.testing.assert_series_equal(
            g._adjacency.sort_index(),
            w.to_adjlist().set_index(["focal", "neighbor"])["weight"].sort_index(),
            check_index_type=False,
            check_dtype=False,
        )
        w_exp = g.to_W()
        # assert w.neighbors == w_exp.neighbors
        assert w.weights == w_exp.weights

        w.transform = "r"
        g_rowwise = graph.Graph.from_W(w)
        pd.testing.assert_series_equal(
            g_rowwise._adjacency.sort_index(),
            w.to_adjlist().set_index(["focal", "neighbor"])["weight"].sort_index(),
            check_index_type=False,
            check_dtype=False,
        )
        w_trans = g_rowwise.to_W()
        # assert w.neighbors == w_trans.neighbors
        assert w.weights == w_trans.weights

        diag = weights.fill_diagonal(w)
        g_diag = graph.Graph.from_W(diag)
        pd.testing.assert_series_equal(
            g_diag._adjacency.sort_index(),
            diag.to_adjlist().set_index(["focal", "neighbor"])["weight"].sort_index(),
            check_index_type=False,
            check_dtype=False,
        )
        w_diag = g_diag.to_W()
        # assert diag.neighbors == W_diag.neighbors
        assert diag.weights == w_diag.weights

        w = weights.W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        g_isolate = graph.Graph.from_W(w)
        pd.testing.assert_series_equal(
            g_isolate._adjacency.sort_index(),
            w.to_adjlist().set_index(["focal", "neighbor"])["weight"].sort_index(),
            check_index_type=False,
            check_dtype=False,
        )
        w_isolate = g_isolate.to_W()
        # assert w.neighbors == w_isolate.neighbors
        assert w.weights == w_isolate.weights

        w = self.g_str_unodered.to_W()
        assert w.id_order_set
        np.testing.assert_array_equal(w.id_order, self.letters[:10])

    def test_from_sparse(self):
        row = np.array([0, 0, 1, 2, 3, 3])
        col = np.array([1, 3, 3, 2, 1, 3])
        data = np.array([0.1, 0.5, 0.9, 0, 0.3, 0.1])
        sp = sparse.coo_array((data, (row, col)), shape=(4, 4))
        g = graph.Graph.from_sparse(sp)
        expected = graph.Graph.from_arrays(row, col, data)
        assert g == expected, "sparse constructor does not match arrays constructor"

        g = graph.Graph.from_sparse(sp.tocsr())
        assert g == expected, "csc input does not match coo input"

        g = graph.Graph.from_sparse(sp.tocsc())
        assert g == expected, "csr input does not match coo input"

        ids = ["zero", "one", "two", "three"]

        g_named = graph.Graph.from_sparse(
            sp,
            ids=ids,
        )

        expected = graph.Graph.from_arrays(
            ["zero", "zero", "one", "two", "three", "three"],
            ["one", "three", "three", "two", "one", "three"],
            data,
        )
        assert g_named == expected

        g = graph.Graph.from_sparse(
            sp.tocsr(),
            ids=ids,
        )
        assert g == expected, "sparse csr with ids does not match arrays constructor"

        g = graph.Graph.from_sparse(
            sp.tocsc(),
            ids=ids,
        )
        assert g == (expected), "sparse csr with ids does not match arrays constructor"

        dense = np.array(
            [
                [0, 0, 1, 1, 0],
                [0, 0, 1, 1, 0],
                [0, 1, 0, 1, 0],
                [0, 1, 0, 0, 1],
                [0, 1, 0, 1, 0],
            ]
        )
        sp = sparse.csr_array(dense)
        g = graph.Graph.from_sparse(
            sp, ids=["staten_island", "queens", "brooklyn", "manhattan", "bronx"]
        )
        expected = graph.Graph.from_arrays(
            [
                "staten_island",
                "staten_island",
                "queens",
                "queens",
                "brooklyn",
                "brooklyn",
                "manhattan",
                "manhattan",
                "bronx",
                "bronx",
            ],
            [
                "brooklyn",
                "manhattan",
                "brooklyn",
                "manhattan",
                "queens",
                "manhattan",
                "queens",
                "bronx",
                "queens",
                "manhattan",
            ],
            np.ones(10),
        )
        assert (
            g == expected
        ), "sparse csr nybb with ids does not match arrays constructor"
        np.testing.assert_array_equal(g.sparse.todense(), sp.todense())

        with pytest.raises(ValueError, match="The length of ids "):
            graph.Graph.from_sparse(sp, ids=["staten_island", "queens"])

    def test_from_arrays(self):
        focal_ids = np.arange(9)
        neighbor_ids = np.array([1, 2, 5, 4, 5, 8, 7, 8, 7])
        weight = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1])

        g = graph.Graph.from_arrays(focal_ids, neighbor_ids, weight)
        pd.testing.assert_series_equal(
            g._adjacency,
            self.adjacency_int_binary,
            check_index_type=False,
            check_dtype=False,
        )

        focal_ids = np.asarray(list(self.neighbor_dict_str.keys()))
        neighbor_ids = np.asarray(list(self.neighbor_dict_str.values()))

        g = graph.Graph.from_arrays(focal_ids, neighbor_ids, weight)
        pd.testing.assert_series_equal(
            g._adjacency,
            self.adjacency_str_binary,
            check_index_type=False,
            check_dtype=False,
        )

    def test_from_weights_dict(self):
        weights_dict = {
            0: {2: 0.5, 1: 0.5},
            1: {0: 0.5, 3: 0.5},
            2: {
                0: 0.3,
                4: 0.3,
                3: 0.3,
            },
            3: {
                1: 0.3,
                2: 0.3,
                5: 0.3,
            },
            4: {2: 0.5, 5: 0.5},
            5: {3: 0.5, 4: 0.5},
        }
        exp_focal = pd.Index(
            [0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 5, 5], dtype="int64", name="focal"
        )
        exp_neighbor = [2, 1, 0, 3, 0, 4, 3, 1, 2, 5, 2, 5, 3, 4]
        exp_weight = [
            0.5,
            0.5,
            0.5,
            0.5,
            0.3,
            0.3,
            0.3,
            0.3,
            0.3,
            0.3,
            0.5,
            0.5,
            0.5,
            0.5,
        ]
        expected = graph.Graph.from_arrays(exp_focal, exp_neighbor, exp_weight)

        g = graph.Graph.from_weights_dict(weights_dict)
        assert g == expected

    def test_from_dicts(self):
        g = graph.Graph.from_dicts(self.neighbor_dict_int)
        pd.testing.assert_series_equal(
            g._adjacency,
            self.adjacency_int_binary,
            check_dtype=False,
            check_index_type=False,
        )

        g = graph.Graph.from_dicts(self.neighbor_dict_str)
        pd.testing.assert_series_equal(
            g._adjacency,
            self.adjacency_str_binary,
            check_dtype=False,
        )

    @pytest.mark.parametrize("y", [3, 5])
    @pytest.mark.parametrize("id_type", ["int", "str"])
    @pytest.mark.parametrize("rook", [True, False])
    def test_from_dicts_via_w(self, y, id_type, rook):
        w = weights.lat2W(3, y, id_type=id_type, rook=rook)
        g = graph.Graph.from_dicts(w.neighbors, w.weights)
        pd.testing.assert_series_equal(
            g._adjacency.sort_index(),
            w.to_adjlist().set_index(["focal", "neighbor"])["weight"].sort_index(),
            check_index_type=False,
            check_dtype=False,
        )

        w.transform = "r"
        g_rowwise = graph.Graph.from_dicts(w.neighbors, w.weights)
        pd.testing.assert_series_equal(
            g_rowwise._adjacency.sort_index(),
            w.to_adjlist().set_index(["focal", "neighbor"])["weight"].sort_index(),
            check_index_type=False,
            check_dtype=False,
        )

        diag = weights.fill_diagonal(w)
        g_diag = graph.Graph.from_dicts(diag.neighbors, diag.weights)
        pd.testing.assert_series_equal(
            g_diag._adjacency.sort_index(),
            diag.to_adjlist().set_index(["focal", "neighbor"])["weight"].sort_index(),
            check_index_type=False,
            check_dtype=False,
        )

        w = weights.W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        g_isolate = graph.Graph.from_dicts(w.neighbors, w.weights)
        pd.testing.assert_series_equal(
            g_isolate._adjacency.sort_index(),
            w.to_adjlist().set_index(["focal", "neighbor"])["weight"].sort_index(),
            check_index_type=False,
            check_dtype=False,
        )

    def test_neighbors(self):
        expected = {
            0: (0, 1, 3),
            1: (0, 2, 4),
            2: (1, 5),
            3: (0, 4, 6),
            4: (1, 3, 5, 7),
            5: (2, 4, 8),
            6: (3, 7),
            7: (4, 6, 8),
            8: (5, 7),
            9: (),
        }
        assert self.g_int.neighbors == expected

        expected = {
            "a": ("a", "b", "d"),
            "b": ("a", "c", "e"),
            "c": ("b", "f"),
            "d": ("a", "e", "g"),
            "e": ("b", "d", "f", "h"),
            "f": ("c", "e", "i"),
            "g": ("d", "h"),
            "h": ("e", "g", "i"),
            "i": ("f", "h"),
            "j": (),
        }
        assert self.g_str.neighbors == expected

    def test_weights(self):
        expected = {
            0: (1.0, 0.5, 0.5),
            1: (0.3, 0.3, 0.3),
            2: (0.5, 0.5),
            3: (0.3, 0.3, 0.3),
            4: (0.25, 0.25, 0.25, 0.25),
            5: (0.3, 0.3, 0.3),
            6: (0.5, 0.5),
            7: (0.3, 0.3, 0.3),
            8: (0.5, 0.5),
            9: (),
        }
        assert self.g_int.weights == expected

        expected = {
            "a": (1.0, 0.5, 0.5),
            "b": (0.3, 0.3, 0.3),
            "c": (0.5, 0.5),
            "d": (0.3, 0.3, 0.3),
            "e": (0.25, 0.25, 0.25, 0.25),
            "f": (0.3, 0.3, 0.3),
            "g": (0.5, 0.5),
            "h": (0.3, 0.3, 0.3),
            "i": (0.5, 0.5),
            "j": (),
        }
        assert self.g_str.weights == expected

    def test_sparse(self):
        sp = self.g_int.sparse
        expected = np.array(
            [
                [1.0, 0.5, 0.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.3, 0.0, 0.3, 0.0, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0, 0.0],
                [0.3, 0.0, 0.0, 0.0, 0.3, 0.0, 0.3, 0.0, 0.0, 0.0],
                [0.0, 0.25, 0.0, 0.25, 0.0, 0.25, 0.0, 0.25, 0.0, 0.0],
                [0.0, 0.0, 0.3, 0.0, 0.3, 0.0, 0.0, 0.0, 0.3, 0.0],
                [0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.3, 0.0, 0.3, 0.0, 0.3, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.5, 0.0, 0.5, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )
        np.testing.assert_array_equal(sp.todense(), expected)

        sp = self.g_str.sparse
        np.testing.assert_array_equal(sp.todense(), expected)

        sp_old = self.g_int.to_W().sparse.todense()
        np.testing.assert_array_equal(sp.todense(), sp_old)

        sp_old = self.g_str.to_W().sparse.todense()
        np.testing.assert_array_equal(sp.todense(), sp_old)

        # check proper sorting
        nybb = graph.Graph.build_contiguity(self.nybb)
        nybb_expected = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 1, 1, 1],
                [0, 1, 0, 1, 0],
                [0, 1, 1, 0, 1],
                [0, 1, 0, 1, 0],
            ]
        )
        np.testing.assert_array_equal(nybb.sparse.todense(), nybb_expected)

    def test_sparse_roundtrip(self):
        g = graph.Graph(self.adjacency_int_binary)
        sp = g.sparse
        g_sp = graph.Graph.from_sparse(sp, self.index_int)
        assert g == g_sp

        g = graph.Graph(self.adjacency_str_binary)
        sp = g.sparse
        g_sp = graph.Graph.from_sparse(sp, self.index_str)
        assert g == g_sp

    def test_cardinalities(self):
        expected = pd.Series(
            [3, 3, 2, 3, 4, 3, 2, 3, 2, 0],
            index=pd.Index(
                ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
                dtype="object",
                name="focal",
            ),
            name="cardinalities",
        )
        pd.testing.assert_series_equal(self.g_str.cardinalities, expected)

    def test_isolates(self):
        expected = pd.Index(["j"], name="focal")
        pd.testing.assert_index_equal(self.g_str.isolates, expected)

        self.g_str._adjacency.iloc[1] = 0  # zero weight, no isolate
        pd.testing.assert_index_equal(self.g_str.isolates, expected)

        with_additional_zeros = self.g_str.assign_self_weight(0)
        pd.testing.assert_index_equal(with_additional_zeros.isolates, expected)

    def test_n(self):
        assert self.g_int.n == 10
        assert self.g_str.n == 10
        assert graph.Graph(self.adjacency_int_binary).n == 9

    def test_pct_nonzero(self):
        assert self.g_int.pct_nonzero == 26.0
        assert graph.Graph(self.adjacency_int_binary).pct_nonzero == pytest.approx(
            11.1111111111
        )

    def test_nonzero(self):
        assert self.g_int.nonzero == 25
        assert graph.Graph(self.adjacency_int_binary).nonzero == 9

    def test_index_pairs(self):
        focal, neighbor = self.g_str.index_pairs
        exp_focal = pd.Index(
            [
                "a",
                "a",
                "a",
                "b",
                "b",
                "b",
                "c",
                "c",
                "d",
                "d",
                "d",
                "e",
                "e",
                "e",
                "e",
                "f",
                "f",
                "f",
                "g",
                "g",
                "h",
                "h",
                "h",
                "i",
                "i",
                "j",
            ],
            name="focal",
        )
        exp_neighbor = pd.Index(
            [
                "a",
                "b",
                "d",
                "a",
                "c",
                "e",
                "b",
                "f",
                "a",
                "e",
                "g",
                "b",
                "d",
                "f",
                "h",
                "c",
                "e",
                "i",
                "d",
                "h",
                "e",
                "g",
                "i",
                "f",
                "h",
                "j",
            ],
            name="neighbor",
        )
        pd.testing.assert_index_equal(exp_focal, focal)
        pd.testing.assert_index_equal(exp_neighbor, neighbor)

    def test_transform_r(self):
        expected_w = [
            0.5,
            0.25,
            0.25,
            0.33333333,
            0.33333333,
            0.33333333,
            0.5,
            0.5,
            0.33333333,
            0.33333333,
            0.33333333,
            0.25,
            0.25,
            0.25,
            0.25,
            0.33333333,
            0.33333333,
            0.33333333,
            0.5,
            0.5,
            0.33333333,
            0.33333333,
            0.33333333,
            0.5,
            0.5,
            0.0,
        ]
        exp = self.g_int.adjacency
        exp.iloc[:] = expected_w
        expected = graph.Graph(exp)
        assert self.g_int.transform("r") == expected
        assert self.g_int.transform("r").transformation == "R"
        assert self.g_int.transform("R") == expected

        w = self.g_int.to_W()
        w.transform = "r"
        g_from_w = graph.Graph.from_W(w)
        assert g_from_w == self.g_int.transform("r")

    def test_transform_b(self):
        expected_w = [
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            0,
        ]
        exp = self.g_int.adjacency
        exp.iloc[:] = expected_w
        expected = graph.Graph(exp)
        assert self.g_int.transform("b") == expected
        assert self.g_int.transform("b").transformation == "B"
        assert self.g_int.transform("B") == expected

        w = self.g_int.to_W()
        w.transform = "b"
        g_from_w = graph.Graph.from_W(w)
        assert g_from_w == self.g_int.transform("b")

    def test_transform_d(self):
        expected_w = [
            0.10416667,
            0.05208333,
            0.05208333,
            0.03125,
            0.03125,
            0.03125,
            0.05208333,
            0.05208333,
            0.03125,
            0.03125,
            0.03125,
            0.02604167,
            0.02604167,
            0.02604167,
            0.02604167,
            0.03125,
            0.03125,
            0.03125,
            0.05208333,
            0.05208333,
            0.03125,
            0.03125,
            0.03125,
            0.05208333,
            0.05208333,
            0.0,
        ]
        exp = self.g_int.adjacency
        exp.iloc[:] = expected_w
        expected = graph.Graph(exp)
        assert self.g_int.transform("d") == expected
        assert self.g_int.transform("d").transformation == "D"
        assert self.g_int.transform("D") == expected
        assert self.g_int.transform("D")._adjacency.sum() == pytest.approx(1)

        w = self.g_int.to_W()
        w.transform = "d"
        g_from_w = graph.Graph.from_W(w)
        assert g_from_w == self.g_int.transform("d")

    def test_transform_v(self):
        expected_w = [
            0.55154388,
            0.27577194,
            0.27577194,
            0.39000042,
            0.39000042,
            0.39000042,
            0.47765102,
            0.47765102,
            0.39000042,
            0.39000042,
            0.39000042,
            0.33775027,
            0.33775027,
            0.33775027,
            0.33775027,
            0.39000042,
            0.39000042,
            0.39000042,
            0.47765102,
            0.47765102,
            0.39000042,
            0.39000042,
            0.39000042,
            0.47765102,
            0.47765102,
            0.0,
        ]
        exp = self.g_int.adjacency
        exp.iloc[:] = expected_w
        expected = graph.Graph(exp)
        assert self.g_int.transform("v") == expected
        assert self.g_int.transform("v").transformation == "V"
        assert self.g_int.transform("V") == expected

        w = self.g_int.to_W()
        w.transform = "v"
        g_from_w = graph.Graph.from_W(w)
        assert g_from_w == self.g_int.transform("v")

    def test_transform(self):
        # do not transform if transformation == current transformation
        binary = self.g_int.transform("b")
        fast_tracked = binary.transform("b")
        assert binary == fast_tracked

        with pytest.raises(ValueError, match="Transformation 'X' is not"):
            self.g_int.transform("x")

    def test_transform_callable(self):
        contig = graph.Graph.build_contiguity(self.nybb)
        trans = contig.transform(lambda x: x * 10)
        assert trans.transformation == "C"
        assert trans.adjacency.sum() == 100

    def test_asymmetry(self):
        neighbors = {
            "a": ["b", "c", "d"],
            "b": ["b", "c", "d"],
            "c": ["a", "b"],
            "d": ["a", "b"],
        }
        weights_d = {"a": [1, 0.5, 1], "b": [1, 1, 1], "c": [1, 1], "d": [1, 1]}
        g = graph.Graph.from_dicts(neighbors, weights_d)
        intrinsic = pd.Series(
            ["b", "c", "a", "a"],
            index=pd.Index(["a", "a", "b", "c"], name="focal"),
            name="neighbor",
        )
        pd.testing.assert_series_equal(intrinsic, g.asymmetry())

        boolean = pd.Series(
            ["b", "a"],
            index=pd.Index(["a", "b"], name="focal"),
            name="neighbor",
        )
        pd.testing.assert_series_equal(boolean, g.asymmetry(intrinsic=False))

        empty = pd.Series(
            index=pd.Index([], name="focal"),
            name="neighbor",
            dtype="int64",
        )

        pd.testing.assert_series_equal(self.g_int.asymmetry(False), empty)

    def test_parquet(self):
        pytest.importorskip("pyarrow")

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "g_int.parquet")
            self.g_int.to_parquet(path)
            g_int = graph.read_parquet(path)
            assert self.g_int == g_int

            path = os.path.join(tmpdir, "g_str.parquet")
            self.g_str.to_parquet(path)
            g_str = graph.read_parquet(path)
            assert self.g_str == g_str

            row_wise = self.g_str.transform("r")
            path = os.path.join(tmpdir, "row.parquet")
            row_wise.to_parquet(path)
            row_read = graph.read_parquet(path)
            assert row_wise == row_read
            assert row_read.transformation == "R"

            path = os.path.join(tmpdir, "pandas.parquet")
            self.g_str._adjacency.to_frame().to_parquet(path)
            g_pandas = graph.read_parquet(path)
            assert self.g_str == g_pandas

    def test_gal(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "g_int.gal")
            g_int = self.g_int.transform("b")
            g_int.to_gal(path)
            g_int_ = graph.read_gal(path)
            assert g_int == g_int_

            path = os.path.join(tmpdir, "g_str.gal")
            g_str = self.g_str.transform("b")
            g_str.to_gal(path)
            g_str_ = graph.read_gal(path)
            assert g_str == g_str_

    def test_gwt(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "g_int.gwt")
            self.g_int.to_gwt(path)
            g_int = graph.read_gwt(path)
            assert self.g_int == g_int

            path = os.path.join(tmpdir, "g_str.gwt")
            self.g_str.to_gwt(path)
            g_str = graph.read_gwt(path)
            assert self.g_str == g_str

    def test_getitem(self):
        expected = pd.Series(
            [1, 0.5, 0.5],
            index=pd.Index(["a", "b", "d"], name="neighbor"),
            name="weight",
        )
        pd.testing.assert_series_equal(expected, self.g_str["a"])

        expected = pd.Series(
            [1, 0.5, 0.5],
            index=pd.Index([0, 1, 3], name="neighbor"),
            name="weight",
        )
        pd.testing.assert_series_equal(expected, self.g_int[0])

        # isolate
        expected = pd.Series(
            [],
            index=pd.Index([], name="neighbor"),
            name="weight",
        )
        pd.testing.assert_series_equal(expected, self.g_str["j"])

    def test_lag(self):
        expected = np.array([4.0, 2.7, 4.0, 3.9, 5.0, 5.1, 6.0, 6.3, 7.0, 0.0])
        lag = self.g_str.lag(list(range(1, 11)))

        np.testing.assert_allclose(expected, lag)

        with pytest.raises(ValueError, match="The length of `y`"):
            self.g_str.lag(list(range(1, 15)))

    @pytest.mark.skipif(
        Version(scipy_version) < Version("1.12.0"),
        reason="sparse matrix power requires scipy>=1.12.0",
    )
    def test_higher_order(self):
        cont = graph.Graph.build_contiguity(self.nybb)
        k2 = cont.higher_order(2)
        expected = graph.Graph.from_arrays(
            self.nybb.index,
            ["Staten Island", "Queens", "Bronx", "Manhattan", "Brooklyn"],
            [0, 0, 1, 0, 1],
        )
        assert k2 == expected

        diagonal = cont.higher_order(2, diagonal=True)
        expected = graph.Graph.from_arrays(
            [
                "Staten Island",
                "Queens",
                "Brooklyn",
                "Brooklyn",
                "Manhattan",
                "Bronx",
                "Bronx",
            ],
            [
                "Staten Island",
                "Queens",
                "Brooklyn",
                "Bronx",
                "Manhattan",
                "Brooklyn",
                "Bronx",
            ],
            [0, 1, 1, 1, 1, 1, 1],
        )
        assert diagonal == expected

        shortest_false = cont.higher_order(2, shortest_path=False)
        expected = graph.Graph.from_arrays(
            [
                "Staten Island",
                "Queens",
                "Queens",
                "Queens",
                "Brooklyn",
                "Brooklyn",
                "Brooklyn",
                "Manhattan",
                "Manhattan",
                "Manhattan",
                "Bronx",
                "Bronx",
                "Bronx",
            ],
            [
                "Staten Island",
                "Brooklyn",
                "Manhattan",
                "Bronx",
                "Queens",
                "Manhattan",
                "Bronx",
                "Queens",
                "Brooklyn",
                "Bronx",
                "Queens",
                "Brooklyn",
                "Manhattan",
            ],
            [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        )
        assert shortest_false == expected

        lower = cont.higher_order(2, lower_order=True)
        assert lower == expected

    @pytest.mark.skipif(
        Version(scipy_version) < Version("1.12.0"),
        reason="sparse matrix power requires scipy>=1.12.0",
    )
    def test_higher_order_inclusive(self):  # GH738
        contig = graph.Graph.from_arrays(
            [0, 1, 2, 3, 3, 4, 4], [0, 3, 4, 1, 4, 2, 3], [0, 1, 1, 1, 1, 1, 1]
        )
        assert len(contig) == 6
        higher = contig.higher_order(2, lower_order=True)
        assert len(higher) == 10
        assert contig < higher

    def test_n_components(self):
        nybb = graph.Graph.build_contiguity(self.nybb)
        assert nybb.n_components == 2

        nybb = graph.Graph.build_knn(self.nybb.set_geometry(self.nybb.centroid), k=2)
        assert nybb.n_components == 1

    def test_component_labels(self):
        nybb = graph.Graph.build_contiguity(self.nybb)
        expected = pd.Series(
            [0, 1, 1, 1, 1],
            index=pd.Index(self.nybb.index.values, name="focal"),
            dtype=int,
            name="component labels",
        )
        pd.testing.assert_series_equal(
            expected, nybb.component_labels, check_dtype=False
        )

    def test_eliminate_zeros(self):
        nybb = graph.Graph.build_contiguity(self.nybb)
        adj = nybb._adjacency.copy()
        adj["Bronx", "Queens"] = 0
        adj["Queens", "Manhattan"] = 0
        adj["Queens", "Queens"] = 0
        with_zero = graph.Graph(adj)
        expected = adj.drop(
            [("Bronx", "Queens"), ("Queens", "Manhattan"), ("Queens", "Queens")]
        )
        pd.testing.assert_series_equal(with_zero.eliminate_zeros()._adjacency, expected)

    def test_subgraph(self):
        knn = graph.Graph.build_knn(self.nybb.set_geometry(self.nybb.centroid), k=2)
        sub = knn.subgraph(["Staten Island", "Bronx", "Brooklyn"])
        assert sub < knn
        expected = pd.Series(
            [1, 0, 0],
            name="weight",
            index=pd.MultiIndex.from_arrays(
                [
                    ["Staten Island", "Bronx", "Brooklyn"],
                    ["Brooklyn", "Bronx", "Brooklyn"],
                ],
                names=["focal", "neighbor"],
            ),
        )
        pd.testing.assert_series_equal(expected, sub._adjacency, check_dtype=False)

    def test_assign_self_weight(self):
        contig = graph.Graph.build_contiguity(self.nybb)
        diag = contig.assign_self_weight()
        assert len(diag._adjacency) == 15
        assert diag._adjacency.sum() == 15

        diag_array = contig.assign_self_weight([2, 3, 4, 5, 6])
        assert len(diag_array._adjacency) == 15
        assert diag_array._adjacency.sum() == 30

        for i, val in enumerate(range(2, 7)):
            assert (
                diag_array._adjacency[(contig.unique_ids[i], contig.unique_ids[i])]
                == val
            )

    def test_apply(self):
        contig = graph.Graph.build_contiguity(self.nybb)

        # pandas built-in
        expected = pd.Series(
            [1.62382200e09, 3.76087588e09, 3.68168493e09, 6.16961834e09, 3.68168493e09],
            index=pd.Index(
                ["Staten Island", "Queens", "Brooklyn", "Manhattan", "Bronx"],
                name="focal",
            ),
        )
        pd.testing.assert_series_equal(contig.apply(self.nybb.area, "sum"), expected)

        # numpy
        expected = pd.Series(
            [1.62382200e09, 1.18692629e09, 1.84084247e09, 1.93747835e09, 1.84084247e09],
            index=pd.Index(
                ["Staten Island", "Queens", "Brooklyn", "Manhattan", "Bronx"],
                name="focal",
            ),
        )
        pd.testing.assert_series_equal(
            contig.apply(self.nybb.area, np.median), expected
        )

        # lambda over geometry
        expected = pd.Series(
            [2.06271959e09, 6.68788190e09, 7.57087991e09, 8.78957337e09, 7.57087991e09],
            index=pd.Index(
                ["Staten Island", "Queens", "Brooklyn", "Manhattan", "Bronx"],
                name="focal",
            ),
        )
        pd.testing.assert_series_equal(
            contig.apply(self.nybb.geometry, lambda x: x.unary_union.convex_hull.area),
            expected,
        )

        # reduction over a dataframe
        expected = pd.DataFrame(
            [
                [3.30470010e05, 1.62381982e09],
                [1.56477261e06, 3.76087473e09],
                [1.25564314e06, 3.68168433e09],
                [2.10181756e06, 6.16961599e09],
                [1.25564314e06, 3.68168433e09],
            ],
            columns=["Shape_Leng", "Shape_Area"],
            index=pd.Index(
                ["Staten Island", "Queens", "Brooklyn", "Manhattan", "Bronx"],
                name="focal",
            ),
        )
        pd.testing.assert_frame_equal(
            contig.apply(
                self.nybb, lambda x: x[["Shape_Leng", "Shape_Area"]].sum(axis=0)
            ),
            expected,
        )

        # 1D array input
        expected = pd.Series(
            [1.62382200e09, 3.76087588e09, 3.68168493e09, 6.16961834e09, 3.68168493e09],
            index=pd.Index(
                ["Staten Island", "Queens", "Brooklyn", "Manhattan", "Bronx"],
                name="focal",
            ),
        )
        pd.testing.assert_series_equal(
            contig.apply(self.nybb.area.values, "sum"), expected
        )

        # 2D array input
        expected = pd.DataFrame(
            [
                [3.30470010e05, 1.62381982e09],
                [1.56477261e06, 3.76087473e09],
                [1.25564314e06, 3.68168433e09],
                [2.10181756e06, 6.16961599e09],
                [1.25564314e06, 3.68168433e09],
            ],
            index=pd.Index(
                ["Staten Island", "Queens", "Brooklyn", "Manhattan", "Bronx"],
                name="focal",
            ),
        )
        pd.testing.assert_frame_equal(
            contig.apply(
                self.nybb[["Shape_Leng", "Shape_Area"]].values,
                lambda x: x.sum(axis=0),
            ),
            expected,
        )

    def test_aggregate(self):
        contig = graph.Graph.build_contiguity(self.nybb)
        expected = pd.Series(
            [1.0, 20.08553692, 7.3890561, 20.08553692, 7.3890561],
            index=pd.Index(
                ["Staten Island", "Queens", "Brooklyn", "Manhattan", "Bronx"],
                name="focal",
            ),
            name="weight",
        )
        pd.testing.assert_series_equal(
            contig.aggregate(lambda x: np.exp(np.sum(x))),
            expected,
        )

    def test_describe(self):
        contig = graph.Graph.build_knn(self.guerry.geometry.centroid, k=5)
        y = self.guerry.geometry.area
        stats = contig.describe(y)
        pd.testing.assert_series_equal(
            stats["count"],
            contig.cardinalities,
            check_names=False,
            check_dtype=False,
        )
        pd.testing.assert_series_equal(
            stats["sum"],
            pd.Series(contig.lag(y), index=contig.unique_ids),
            check_names=False,
        )
        r_contig = contig.transform("R")
        pd.testing.assert_series_equal(
            stats["mean"],
            pd.Series(r_contig.lag(y), index=contig.unique_ids),
            check_names=False,
        )
        ## compute only some statistics
        specific_stats = contig.describe(y, statistics=["count", "sum", "mean"])
        ## assert only the specified values are computed
        assert list(specific_stats.columns) == ["count", "sum", "mean"]

        pd.testing.assert_frame_equal(
            specific_stats[["count", "sum", "mean"]], stats[["count", "sum", "mean"]]
        )

        percentile_stats = contig.describe(y, q=(25, 75))

        for i in contig.unique_ids:
            neigh_vals = y[contig[i].index.values]
            low, high = neigh_vals.describe()[["25%", "75%"]]
            neigh_vals = neigh_vals[(low <= neigh_vals) & (neigh_vals <= high)]
            expected = neigh_vals.describe()[["count", "mean", "std", "min", "max"]]
            res = percentile_stats.loc[i][["count", "mean", "std", "min", "max"]]
            pd.testing.assert_series_equal(res, expected, check_names=False)

        ## test NA equivalence between filtration and pandas
        nan_areas = y.copy()
        nan_areas.iloc[range(0, len(y), 3),] = np.nan
        res1 = contig.describe(y, statistics=["count"])["count"]
        res2 = contig.describe(y, statistics=["count"], q=(0, 100))["count"]
        pd.testing.assert_series_equal(res1, res2)

        # test with isolates and string index
        nybb_contig = graph.Graph.build_contiguity(self.nybb, rook=False)
        stats = nybb_contig.describe(
            self.nybb.geometry.area, statistics=["count", "sum", "mode"]
        )
        ## all isolate values should be nan
        assert stats.loc["Staten Island"].isna().all()

        # for easier comparison and na has already been checked.
        stats = stats.fillna(0)

        pd.testing.assert_series_equal(
            stats["sum"],
            pd.Series(nybb_contig.lag(self.nybb.geometry.area), index=self.nybb.index),
            check_names=False,
        )

        pd.testing.assert_series_equal(
            stats["count"].sort_index(),
            nybb_contig.cardinalities.sort_index(),
            check_dtype=False,
            check_names=False,
        )

        y = self.nybb.geometry.area
        for i in nybb_contig.unique_ids:
            neigh_vals = y.loc[nybb_contig[i].index.values]
            expected = neigh_vals.mode().iloc[0] if neigh_vals.shape[0] else 0
            res = stats.loc[i]["mode"]
            assert res == expected

        ## test passing ndarray
        stats1 = nybb_contig.describe(self.nybb.geometry.area, statistics=["sum"])[
            "sum"
        ]
        stats2 = nybb_contig.describe(
            self.nybb.geometry.area.values, statistics=["sum"]
        )["sum"]
        pd.testing.assert_series_equal(
            stats1,
            stats2,
            check_dtype=False,
            check_names=False,
        )

        ## test index alignment
        with pytest.raises(
            ValueError, match="The values index is not aligned with the graph index."
        ):
            nybb_contig.describe(self.nybb.geometry.area.reset_index(drop=True))

    def test_summary(self):
        assert isinstance(self.g_int.summary(), graph.GraphSummary)
        assert isinstance(self.g_str.summary(), graph.GraphSummary)
        assert isinstance(self.g_str_unodered.summary(), graph.GraphSummary)



import numpy as np
import pandas as pd


def _lag_spatial(graph, y, categorical=False, ties="raise"):
    """Spatial lag operator

    Constructs spatial lag based on neighbor relations of the graph.


    Parameters
    ----------
    graph : Graph
        libpysal.graph.Graph
    y : array
        numpy array with dimensionality conforming to w
    categorical : bool
        True if y is categorical, False if y is continuous.
    ties : {'raise', 'random', 'tryself'}, optional
        Policy on how to break ties when a focal unit has multiple
        modes for a categorical lag.
        - 'raise': This will raise an exception if ties are
          encountered to alert the user (Default).
        - 'random': modal label ties Will be broken randomly.
        - 'tryself': check if focal label breaks the tie between label
          modes.  If the focal label does not break the modal tie, the
          tie will be be broken randomly. If the focal unit has a
          self-weight, focal label is not used to break any tie,
          rather any tie will be broken randomly.


    Returns
    -------
    numpy.array
        array of numeric|categorical values for the spatial lag


    Examples
    --------
    >>> from libpysal.graph._spatial_lag import _lag_spatial
    >>> import numpy as np
    >>> from libpysal.weights.util import lat2W
    >>> from libpysal.graph import Graph
    >>> graph = Graph.from_W(lat2W(3,3))
    >>> y = np.arange(9)
    >>> _lag_spatial(graph, y)
    array([ 4.,  6.,  6., 10., 16., 14., 10., 18., 12.])

    Row standardization
    >>> w = lat2W(3,3)
    >>> w.transform = 'r'
    >>> graph = Graph.from_W(w)
    >>> y = np.arange(9)
    >>> _lag_spatial(graph, y)
    array([2.        , 2.        , 3.        , 3.33333333, 4.        ,
           4.66666667, 5.        , 6.        , 6.        ])


    Categorical Lag (no ties)
    >>> y = np.array([*'ababcbcbc'])
    >>> _lag_spatial(graph, y, categorical=True)
    array(['b', 'a', 'b', 'c', 'b', 'c', 'b', 'c', 'b'], dtype=object)

    Handling ties
    >>> y[3] = 'a'
    >>> np.random.seed(12345)
    >>> _lag_spatial(graph, y, categorical=True, ties='random')
    array(['a', 'a', 'b', 'c', 'b', 'c', 'b', 'c', 'b'], dtype=object)
    >>> _lag_spatial(graph, y, categorical=True, ties='random')
    array(['b', 'a', 'b', 'c', 'b', 'c', 'b', 'c', 'b'], dtype=object)
    >>> _lag_spatial(graph, y, categorical=True, ties='tryself')
    array(['a', 'a', 'b', 'c', 'b', 'c', 'a', 'c', 'b'], dtype=object)

    """
    sp = graph.sparse
    if len(y) != sp.shape[0]:
        raise ValueError(
            "The length of `y` needs to match the number of observations "
            f"in Graph. Expected {sp.shape[0]}, got {len(y)}."
        )

    # coerce list to array
    if isinstance(y, list):
        y = np.array(y)

    if (
        isinstance(y.dtype, pd.CategoricalDtype)
        or pd.api.types.is_object_dtype(y.dtype)
        or pd.api.types.is_bool_dtype(y.dtype)
        or pd.api.types.is_string_dtype(y.dtype)
    ):
        categorical = True
    if categorical:
        if isinstance(y, np.ndarray):
            y = pd.Series(y, index=graph.unique_ids)

        df = pd.DataFrame(data=graph.adjacency)
        df["neighbor_label"] = y.loc[graph.adjacency.index.get_level_values(1)].values
        df["own_label"] = y.loc[graph.adjacency.index.get_level_values(0)].values
        df["neighbor_idx"] = df.index.get_level_values(1)
        df["focal_idx"] = df.index.get_level_values(0)
        gb = df.groupby(["focal", "neighbor_label"]).count().groupby(level="focal")
        n_ties = gb.apply(_check_ties).sum()
        if n_ties and ties == "raise":
            raise ValueError(
                f"There are {n_ties} ties that must be broken "
                f"to define the categorical "
                "spatial lag for these observations. To address this "
                "issue, consider setting `ties='tryself'` "
                "or `ties='random'` or consult the documentation "
                "about ties and the categorical spatial lag."
            )
        # either there are ties and random|tryself specified or
        # there are no ties
        gb = df.groupby(by=["focal"])
        if ties == "random" or ties == "raise":
            return gb.apply(_get_categorical_lag).values
        elif ties == "tryself" or ties == "raise":
            return gb.apply(_get_categorical_lag, ties="tryself").values
        else:
            raise ValueError(
                f"Received option ties='{ties}', but only options "
                "'raise','random','tryself' are supported."
            )

    return sp @ y


def _check_ties(focal):
    """Reduction to determine if a focal unit has multiple modes for neighbor labels.

    Parameters
    ----------
    focal: row from pandas Dataframe
          Data is a Graph with an additional column having the labels for the neighbors

    Returns
    -------
    bool
    """

    max_count = focal.weight.max()
    return (focal.weight == max_count).sum() > 1


def _get_categorical_lag(focal, ties="random"):
    """Reduction to determine categorical spatial lag for a focal unit.

    Parameters
    ----------
    focal: row from pandas Dataframe
          Data is a Graph with an additional column having the labels for the neighbors

    ties : {'raise', 'random', 'tryself'}, optional
        Policy on how to break ties when a focal unit has multiple
        modes for a categorical lag.
        - 'raise': This will raise an exception if ties are
          encountered to alert the user (Default).
        - 'random': Will break ties randomly.
        - 'tryself': check if focal label breaks the tie between label
          modes.  If the focal label does not break the modal tie, the
          tie will be be broken randomly. If the focal unit has a
          self-weight, focal label is not used to break any tie,
          rather any tie will be broken randomly.


    Returns
    -------
    str|int|float:
      Label for the value of the categorical lag
    """
    self_weight = focal.focal_idx.values[0] in focal.neighbor_idx.values
    labels, counts = np.unique(focal.neighbor_label, return_counts=True)
    node_label = labels[counts == counts.max()]
    if ties == "random" or (ties == "tryself" and self_weight):
        return np.random.choice(node_label, 1)[0]
    elif ties == "tryself" and not self_weight:
        self_label = focal.own_label.values[0]
        if self_label in node_label:  # focal breaks tie
            return self_label
        else:
            return np.random.choice(node_label, 1)[0]



import numpy as np

from ....common import requires as _requires
from ....examples import get_path
from ....io import geotable as pdio
from ... import ops as GIS  # noqa: N812
from ...shapes import Polygon


class TestTabular:
    def setup_method(self):
        import pandas as pd

        self.columbus = pdio.read_files(get_path("columbus.shp"))
        grid = [
            Polygon([(0, 0), (0, 1), (1, 1), (1, 0)]),
            Polygon([(0, 1), (0, 2), (1, 2), (1, 1)]),
            Polygon([(1, 2), (2, 2), (2, 1), (1, 1)]),
            Polygon([(1, 1), (2, 1), (2, 0), (1, 0)]),
        ]
        regime = [0, 0, 1, 1]
        ids = list(range(4))
        data = np.array((regime, ids)).T
        self.exdf = pd.DataFrame(data, columns=["regime", "ids"])
        self.exdf["geometry"] = grid

    @_requires("geopandas")
    def test_round_trip(self):
        import geopandas as gpd
        import pandas as pd

        geodf = GIS.tabular.to_gdf(self.columbus)
        assert isinstance(geodf, gpd.GeoDataFrame)
        new_df = GIS.tabular.to_df(geodf)
        assert isinstance(new_df, pd.DataFrame)
        for new, old in zip(new_df.geometry, self.columbus.geometry, strict=True):
            assert new == old

    def test_spatial_join(self):
        pass

    def test_spatial_overlay(self):
        pass

    def test_dissolve(self):
        out = GIS.tabular.dissolve(self.exdf, by="regime")
        assert out[0].area == 2.0
        assert out[1].area == 2.0

        answer_vertices0 = {(0, 0), (0, 1), (0, 2), (1, 2), (1, 1), (1, 0), (0, 0)}
        answer_vertices1 = {(2, 1), (2, 0), (1, 0), (1, 1), (1, 2), (2, 2), (2, 1)}

        s0 = {tuple(map(int, t)) for t in out[0].vertices}
        s1 = {tuple(map(int, t)) for t in out[1].vertices}

        assert s0 == answer_vertices0
        assert s1 == answer_vertices1

    def test_clip(self):
        pass

    def test_erase(self):
        pass

    def test_union(self):
        new_geom = GIS.tabular.union(self.exdf)
        assert new_geom.area == 4

    def test_intersection(self):
        pass

    def test_symmetric_difference(self):
        pass

    def test_difference(self):
        pass



import copy
import numbers
import os
from collections import defaultdict
from itertools import tee
from warnings import warn

import numpy as np

# ruff: noqa: N802, N803
import scipy
import scipy.spatial
from packaging.version import Version
from scipy import sparse
from scipy.spatial import KDTree

from ..common import requires
from ..io.fileio import FileIO
from .set_operations import w_subset
from .weights import WSP, W

try:
    import geopandas as gpd

    GPD_013 = Version(gpd.__version__) >= Version("0.13.0")
except ImportError:
    warn("geopandas not available. Some functionality will be disabled.", stacklevel=2)

try:
    from shapely.geometry.base import BaseGeometry

    HAS_SHAPELY = True
except ImportError:
    HAS_SHAPELY = False

__all__ = [
    "lat2W",
    "block_weights",
    "comb",
    "order",
    "higher_order",
    "shimbel",
    "remap_ids",
    "full2W",
    "full",
    "WSP2W",
    "insert_diagonal",
    "fill_diagonal",
    "get_ids",
    "get_points_array_from_shapefile",
    "min_threshold_distance",
    "lat2SW",
    "w_local_cluster",
    "higher_order_sp",
    "hexLat2W",
    "neighbor_equality",
    "attach_islands",
    "nonplanar_neighbors",
    "fuzzy_contiguity",
]


KDTREE_TYPES = [scipy.spatial.KDTree, scipy.spatial.cKDTree]


def hexLat2W(nrows=5, ncols=5, **kwargs):
    """
    Create a W object for a hexagonal lattice.

    Parameters
    ----------
    nrows      : int
                 number of rows
    ncols      : int
                 number of columns
    **kwargs   : keyword arguments
                 optional arguments for :class:`pysal.weights.W`

    Returns
    -------
    w : W
        instance of spatial weights class W

    Notes
    -----
    Observations are row ordered: first k observations are in row 0,
    next k in row 1, and so on.

    Construction is based on shifting every other column of a regular lattice
    down 1/2 of a cell.

    Examples
    --------
    >>> from libpysal.weights import lat2W, hexLat2W
    >>> w = lat2W()
    >>> w.neighbors[1]
    [0, 6, 2]
    >>> w.neighbors[21]
    [16, 20, 22]
    >>> wh = hexLat2W()
    >>> wh.neighbors[1]
    [0, 6, 2, 5, 7]
    >>> wh.neighbors[21]
    [16, 20, 22]
    """

    if nrows == 1 or ncols == 1:
        print("Hexagon lattice requires at least 2 rows and columns")
        print("Returning a linear contiguity structure")
        return lat2W(nrows, ncols)

    n = nrows * ncols
    rid = [i // ncols for i in range(n)]
    cid = [i % ncols for i in range(n)]
    r1 = nrows - 1
    c1 = ncols - 1

    w = lat2W(nrows, ncols).neighbors
    for i in range(n):
        odd = cid[i] % 2
        if odd:
            if rid[i] < r1:  # odd col index above last row
                # new sw neighbor
                if cid[i] > 0:
                    j = i + ncols - 1
                    w[i] = w.get(i, []) + [j]
                # new se neighbor
                if cid[i] < c1:
                    j = i + ncols + 1
                    w[i] = w.get(i, []) + [j]

        else:  # even col
            # nw
            jnw = [i - ncols - 1]
            # ne
            jne = [i - ncols + 1]
            if rid[i] > 0:
                w[i]
                if cid[i] == 0:
                    w[i] = w.get(i, []) + jne
                elif cid[i] == c1:
                    w[i] = w.get(i, []) + jnw
                else:
                    w[i] = w.get(i, []) + jne
                    w[i] = w.get(i, []) + jnw

    return W(w, **kwargs)


def lat2W(nrows=5, ncols=5, rook=True, id_type="int", **kwargs):
    """
    Create a W object for a regular lattice.

    Parameters
    ----------
    nrows      : int
                 number of rows
    ncols      : int
                 number of columns
    rook       : boolean
                 type of contiguity. Default is rook. For queen, rook =False
    id_type    : string
                 string defining the type of IDs to use in the final W object;
                 options are 'int' (0, 1, 2 ...; default), 'float' (0.0,
                 1.0, 2.0, ...) and 'string' ('id0', 'id1', 'id2', ...)
    **kwargs   : keyword arguments
                 optional arguments for :class:`pysal.weights.W`

    Returns
    -------
    w : W
        instance of spatial weights class W

    Notes
    -----
    Observations are row ordered: first k observations are in row 0,
    next k in row 1, and so on.

    Examples
    --------
    >>> from libpysal.weights import lat2W
    >>> w9 = lat2W(3,3)
    >>> "%.3f"%w9.pct_nonzero
    '29.630'
    >>> w9[0] == {1: 1.0, 3: 1.0}
    True
    >>> w9[3] == {0: 1.0, 4: 1.0, 6: 1.0}
    True
    """
    n = nrows * ncols
    r1 = nrows - 1
    c1 = ncols - 1
    rid = [i // ncols for i in range(n)]  # must be floor!
    cid = [i % ncols for i in range(n)]
    w = {}
    r = below = 0
    for i in range(n - 1):
        if rid[i] < r1:
            below = rid[i] + 1
            r = below * ncols + cid[i]
            w[i] = w.get(i, []) + [r]
            w[r] = w.get(r, []) + [i]
        if cid[i] < c1:
            right = cid[i] + 1
            c = rid[i] * ncols + right
            w[i] = w.get(i, []) + [c]
            w[c] = w.get(c, []) + [i]
        if not rook:
            # southeast bishop
            if cid[i] < c1 and rid[i] < r1:
                r = (rid[i] + 1) * ncols + 1 + cid[i]
                w[i] = w.get(i, []) + [r]
                w[r] = w.get(r, []) + [i]
            # southwest bishop
            if cid[i] > 0 and rid[i] < r1:
                r = (rid[i] + 1) * ncols - 1 + cid[i]
                w[i] = w.get(i, []) + [r]
                w[r] = w.get(r, []) + [i]

    weights = {}
    for key in w:
        weights[key] = [1.0] * len(w[key])
    ids = list(range(n))
    if id_type == "string":
        ids = ["id" + str(i) for i in ids]
    elif id_type == "float":
        ids = [i * 1.0 for i in ids]
    if id_type == "string" or id_type == "float":
        id_dict = dict(list(zip(list(range(n)), ids, strict=True)))
        alt_w = {}
        alt_weights = {}
        for i in w:
            values = [id_dict[j] for j in w[i]]
            key = id_dict[i]
            alt_w[key] = values
            alt_weights[key] = weights[i]
        w = alt_w
        weights = alt_weights
    return W(w, weights, ids=ids, id_order=ids[:], **kwargs)


def block_weights(regimes, ids=None, sparse=False, **kwargs):
    """
    Construct spatial weights for regime neighbors.

    Block contiguity structures are relevant when defining neighbor relations
    based on membership in a regime. For example, all counties belonging to
    the same state could be defined as neighbors, in an analysis of all
    counties in the US.

    Parameters
    ----------
    regimes     : list, array
                  ids of which regime an observation belongs to
    ids         : list, array
                  Ordered sequence of IDs for the observations
    sparse      : boolean
                  If True return WSP instance
                  If False return W instance
    **kwargs    : keyword arguments
                  optional arguments for :class:`pysal.weights.W`

    Returns
    -------
    W : spatial weights instance

    Examples
    --------
    >>> from libpysal.weights import block_weights
    >>> import numpy as np
    >>> regimes = np.ones(25)
    >>> regimes[range(10,20)] = 2
    >>> regimes[range(21,25)] = 3
    >>> regimes
    array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 2., 2., 2., 2., 2., 2., 2.,
           2., 2., 2., 1., 3., 3., 3., 3.])
    >>> w = block_weights(regimes)
    >>> w.weights[0]
    [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    >>> w.neighbors[0]
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 20]
    >>> regimes = ['n','n','s','s','e','e','w','w','e']
    >>> n = len(regimes)
    >>> w = block_weights(regimes)
    >>> w.neighbors == {0: [1], 1: [0], 2: [3], 3: [2], 4: [5, 8], 5: [4, 8], 6: [7], 7: [6], 8: [4, 5]}
    True
    """  # noqa: E501
    rids = np.unique(regimes)
    neighbors = {}
    npnz = np.nonzero
    regimes = np.array(regimes)
    for rid in rids:
        members = npnz(regimes == rid)[0]
        for member in members:
            neighbors[member] = members[npnz(members != member)[0]].tolist()
    w = W(neighbors, **kwargs)
    if ids is not None:
        w.remap_ids(ids)
    if sparse:
        w = WSP(w.sparse, id_order=ids)
    return w


def comb(items, n=None):
    """
    Combinations of size n taken from items

    Parameters
    ----------
    items : list
            items to be drawn from
    n     : integer
            size of combinations to take from items

    Returns
    -------
    implicit : generator
               combinations of size n taken from items

    Examples
    --------
    >>> x = range(4)
    >>> for c in comb(x, 2):
    ...     print(c)
    ...
    [0, 1]
    [0, 2]
    [0, 3]
    [1, 2]
    [1, 3]
    [2, 3]
    """
    items = list(items)
    if n is None:
        n = len(items)
    for i in list(range(len(items))):
        v = items[i : i + 1]
        if n == 1:
            yield v
        else:
            rest = items[i + 1 :]
            for c in comb(rest, n - 1):
                yield v + c


def order(w, kmax=3):
    """
    Determine the non-redundant order of contiguity up to a specific
    order.

    Parameters
    ----------
    w       : W
              spatial weights object

    kmax    : int
              maximum order of contiguity

    Returns
    -------
    info : dictionary
           observation id is the key, value is a list of contiguity
           orders with a negative 1 in the ith position

    Notes
    -----
    Implements the algorithm in :cite:`Anselin1996b`.

    Examples
    --------
    >>> from libpysal.weights import order, Rook
    >>> import libpysal
    >>> w = Rook.from_shapefile(libpysal.examples.get_path('10740.shp'))

    WARNING: there is one disconnected observation (no neighbors)
    Island id:  [163]
    >>> w3 = order(w, kmax = 3)
    >>> w3[1][0:5]
    [1, -1, 1, 2, 1]
    """

    ids = w.id_order
    info = {}
    for id_ in ids:
        s = [0] * w.n
        s[ids.index(id_)] = -1
        for j in w.neighbors[id_]:
            s[ids.index(j)] = 1
        k = 1
        while k < kmax:
            knext = k + 1
            if s.count(k):
                # get neighbors of order k
                js = [ids[j] for j, val in enumerate(s) if val == k]
                # get first order neighbors for order k neighbors
                for j in js:
                    next_neighbors = w.neighbors[j]
                    for neighbor in next_neighbors:
                        nid = ids.index(neighbor)
                        if s[nid] == 0:
                            s[nid] = knext
            k = knext
        info[id_] = s
    return info


def higher_order(w, k=2, **kwargs):
    """
    Contiguity weights object of order k.

    Parameters
    ----------
    w        : W
               spatial weights object
    k        : int
               order of contiguity
    **kwargs : keyword arguments
               optional arguments for :class:`pysal.weights.W`

    Returns
    -------
    implicit : W
               spatial weights object

    Notes
    -----
    Proper higher order neighbors are returned such that i and j are k-order
    neighbors iff the shortest path from i-j is of length k.

    Examples
    --------
    >>> from libpysal.weights import lat2W, higher_order
    >>> w10 = lat2W(10, 10)
    >>> w10_2 = higher_order(w10, 2)
    >>> w10_2[0] ==  {2: 1.0, 11: 1.0, 20: 1.0}
    True
    >>> w5 = lat2W()
    >>> w5[0] ==  {1: 1.0, 5: 1.0}
    True
    >>> w5[1] == {0: 1.0, 2: 1.0, 6: 1.0}
    True
    >>> w5_2 = higher_order(w5,2)
    >>> w5_2[0] == {10: 1.0, 2: 1.0, 6: 1.0}
    True
    """
    return higher_order_sp(w, k, **kwargs)


def higher_order_sp(
    w, k=2, shortest_path=True, diagonal=False, lower_order=False, **kwargs
):
    """
    Contiguity weights for either a sparse W or W for order k.

    Parameters
    ----------
    w             : W
                            sparse_matrix, spatial weights object or
                            scipy.sparse.csr.csr_instance
    k             : int
                    Order of contiguity
    shortest_path : boolean
                    True: i,j and k-order neighbors if the
                    shortest path for i,j is k.
                    False: i,j are k-order neighbors if there
                    is a path from i,j of length k.
    diagonal      : boolean
                    True:  keep k-order (i,j) joins when i==j
                    False: remove k-order (i,j) joins when i==j
    lower_order   : boolean
                    True: include lower order contiguities
                    False: return only weights of order k
    **kwargs      : keyword arguments
                    optional arguments for :class:`pysal.weights.W`

    Returns
    -------
    wk : W
             WSP, type matches type of w argument

    Examples
    --------
    >>> from libpysal.weights import lat2W, higher_order_sp
    >>> w25 = lat2W(5,5)
    >>> w25.n
    25
    >>> w25[0] == {1: 1.0, 5: 1.0}
    True
    >>> w25_2 = higher_order_sp(w25, 2)
    >>> w25_2[0] == {10: 1.0, 2: 1.0, 6: 1.0}
    True
    >>> w25_2 = higher_order_sp(w25, 2, diagonal=True)
    >>> w25_2[0] ==  {0: 1.0, 10: 1.0, 2: 1.0, 6: 1.0}
    True
    >>> w25_3 = higher_order_sp(w25, 3)
    >>> w25_3[0] == {15: 1.0, 3: 1.0, 11: 1.0, 7: 1.0}
    True
    >>> w25_3 = higher_order_sp(w25, 3, shortest_path=False)
    >>> w25_3[0] == {1: 1.0, 3: 1.0, 5: 1.0, 7: 1.0, 11: 1.0, 15: 1.0}
    True
    >>> w25_3 = higher_order_sp(w25, 3, lower_order=True)
    >>> w25_3[0] == {
    ...     5: 1.0, 7: 1.0, 11: 1.0, 2: 1.0, 15: 1.0, 6: 1.0, 10: 1.0, 1: 1.0, 3: 1.0
    ... }
    True
    """
    id_order = None
    if issubclass(type(w), W) or isinstance(w, W):
        if np.unique(np.hstack(list(w.weights.values()))) == np.array([1.0]):
            id_order = w.id_order
            w = w.sparse
        else:
            raise ValueError("Weights are not binary (0,1)")
    elif scipy.sparse.isspmatrix_csr(w):
        if not np.unique(w.data) == np.array([1.0]):
            raise ValueError(
                "Sparse weights matrix is not binary (0,1) weights matrix."
            )
    else:
        raise TypeError(
            "Weights provided are neither a binary W object nor "
            "a scipy.sparse.csr_matrix"
        )

    if lower_order:
        wk = sum(w**x for x in range(1, k + 1))
        shortest_path = False
    else:
        wk = w**k

    rk, ck = wk.nonzero()
    sk = set(zip(rk, ck, strict=True))

    if shortest_path:
        for j in range(1, k):
            wj = w**j
            rj, cj = wj.nonzero()
            sj = set(zip(rj, cj, strict=True))
            sk.difference_update(sj)

    if not diagonal:
        sk = {(i, j) for i, j in sk if i != j}

    if id_order:
        d = {i: [] for i in id_order}
        for pair in sk:
            k, v = pair
            k = id_order[k]
            v = id_order[v]
            d[k].append(v)
        return W(neighbors=d, **kwargs)
    else:
        d = {}
        for pair in sk:
            k, v = pair
            if k in d:
                d[k].append(v)
            else:
                d[k] = [v]
        return WSP(W(neighbors=d, **kwargs).sparse)


def w_local_cluster(w):
    r"""
    Local clustering coefficients for each unit as a node in a graph.

    Parameters
    ----------
    w   : W
          spatial weights object

    Returns
    -------
    c     : array
            (w.n,1)
            local clustering coefficients

    Notes
    -----
    The local clustering coefficient :math:`c_i` quantifies how close the
    neighbors of observation :math:`i` are to being a clique:

    .. math::

       c_i = | \{w_{j,k}\} |/ (k_i(k_i - 1)): j,k \in N_i

    where :math:`N_i` is the set of neighbors to :math:`i`, :math:`k_i =
    |N_i|` and :math:`\{w_{j,k}\}` is the set of non-zero elements of the
    weights between pairs in :math:`N_i` :cite:`Watts1998`.

    Examples
    --------
    >>> from libpysal.weights import lat2W, w_local_cluster
    >>> w = lat2W(3,3, rook=False)
    >>> w_local_cluster(w)
    array([[1.        ],
           [0.6       ],
           [1.        ],
           [0.6       ],
           [0.42857143],
           [0.6       ],
           [1.        ],
           [0.6       ],
           [1.        ]])
    """

    c = np.zeros((w.n, 1), float)
    w.transformation = "b"
    for i, id_ in enumerate(w.id_order):
        ki = max(w.cardinalities[id_], 1)  # deal with islands
        ni = w.neighbors[id_]
        wi = w_subset(w, ni).full()[0]
        c[i] = wi.sum() / (ki * (ki - 1))
    return c


def shimbel(w):
    """
    Find the Shimbel matrix for first order contiguity matrix.

    Parameters
    ----------
    w     : W
            spatial weights object

    Returns
    -------
    info  : list
            list of lists; one list for each observation which stores
            the shortest order between it and each of the the other observations.

    Examples
    --------
    >>> from libpysal.weights import lat2W, shimbel
    >>> w5 = lat2W()
    >>> w5_shimbel = shimbel(w5)
    >>> w5_shimbel[0][24]
    8
    >>> w5_shimbel[0][0:4]
    [-1, 1, 2, 3]
    """

    info = {}
    ids = w.id_order
    for i in ids:
        s = [0] * w.n
        s[ids.index(i)] = -1
        for j in w.neighbors[i]:
            s[ids.index(j)] = 1
        k = 1
        flag = s.count(0)
        while flag:
            p = -1
            knext = k + 1
            for _ in range(s.count(k)):
                neighbor = s.index(k, p + 1)
                p = neighbor
                next_neighbors = w.neighbors[ids[p]]
                for neighbor in next_neighbors:
                    nid = ids.index(neighbor)
                    if s[nid] == 0:
                        s[nid] = knext
            k = knext
            flag = s.count(0)
        info[i] = s
    return info


def full(w):
    """
    Generate a full numpy array.

    Parameters
    ----------
    w        : W
               spatial weights object

    Returns
    -------
    (fullw, keys) : tuple
                    first element being the full numpy array and second element
                    keys being the ids associated with each row in the array.

    Examples
    --------
    >>> from libpysal.weights import W, full
    >>> neighbors = {'first':['second'],'second':['first','third'],'third':['second']}
    >>> weights = {'first':[1],'second':[1,1],'third':[1]}
    >>> w = W(neighbors, weights)
    >>> wf, ids = full(w)
    >>> wf
    array([[0., 1., 0.],
           [1., 0., 1.],
           [0., 1., 0.]])
    >>> ids
    ['first', 'second', 'third']
    """
    return w.full()


def full2W(m, ids=None, **kwargs):
    """
    Create a PySAL W object from a full array.

    Parameters
    ----------
    m          : array
                 nxn array with the full weights matrix
    ids        : list
                 User ids assumed to be aligned with m
    **kwargs   : keyword arguments
                 optional arguments for :class:`pysal.weights.W`

    Returns
    -------
    w       : W
              PySAL weights object

    Examples
    --------
    >>> from libpysal.weights import full2W
    >>> import numpy as np

    Create an array of zeros

    >>> a = np.zeros((4, 4))

    For loop to fill it with random numbers

    >>> for i in range(len(a)):
    ...     for j in range(len(a[i])):
    ...         if i!=j:
    ...             a[i, j] = np.random.random(1)

    Create W object

    >>> w = full2W(a)
    >>> w.full()[0] == a
    array([[ True,  True,  True,  True],
           [ True,  True,  True,  True],
           [ True,  True,  True,  True],
           [ True,  True,  True,  True]])

    Create list of user ids

    >>> ids = ['myID0', 'myID1', 'myID2', 'myID3']
    >>> w = full2W(a, ids=ids)
    >>> w.full()[0] == a
    array([[ True,  True,  True,  True],
           [ True,  True,  True,  True],
           [ True,  True,  True,  True],
           [ True,  True,  True,  True]])
    """
    if m.shape[0] != m.shape[1]:
        raise ValueError("Your array is not square")
    neighbors, weights = {}, {}
    for i in range(m.shape[0]):
        # for i, row in enumerate(m):
        row = m[i]
        if ids:
            i = ids[i]
        ngh = list(row.nonzero()[0])
        weights[i] = list(row[ngh])
        ngh = list(ngh)
        if ids:
            ngh = [ids[j] for j in ngh]
        neighbors[i] = ngh
    return W(neighbors, weights, id_order=ids, **kwargs)


def WSP2W(wsp, **kwargs):
    """
    Convert a pysal WSP object (thin weights matrix) to a pysal W object.

    Parameters
    ----------
    wsp                     : WSP
                              PySAL sparse weights object
    **kwargs                : keyword arguments
                              optional arguments for :class:`pysal.weights.W`

    Returns
    -------
    w       : W
              PySAL weights object

    Examples
    --------
    >>> from libpysal.weights import lat2W, WSP, WSP2W

    Build a 10x10 scipy.sparse matrix for a rectangular 2x5 region of cells
    (rook contiguity), then construct a PySAL sparse weights object (wsp).

    >>> sp = lat2SW(2, 5)
    >>> wsp = WSP(sp)
    >>> wsp.n
    10
    >>> wsp.sparse[0].todense()
    matrix([[0, 1, 0, 0, 0, 1, 0, 0, 0, 0]], dtype=int8)

    Convert this sparse weights object to a standard PySAL weights object.

    >>> w = WSP2W(wsp)
    >>> w.n
    10
    >>> print(w.full()[0][0])
    [0. 1. 0. 0. 0. 1. 0. 0. 0. 0.]
    """
    data = wsp.sparse.data
    indptr = wsp.sparse.indptr
    id_order = wsp.id_order
    if id_order:
        # replace indices with user IDs
        indices = [id_order[i] for i in wsp.sparse.indices]
    else:
        id_order = list(range(wsp.n))
    neighbors, weights = {}, {}
    start = indptr[0]
    for i in range(wsp.n):
        oid = id_order[i]
        end = indptr[i + 1]
        neighbors[oid] = indices[start:end]
        weights[oid] = data[start:end].tolist()
        start = end
    ids = copy.copy(wsp.id_order)
    w = W(neighbors, weights, ids, **kwargs)
    w._sparse = copy.deepcopy(wsp.sparse)
    w._cache["sparse"] = w._sparse
    return w


def insert_diagonal(w, val=1.0, wsp=False):
    warn("This function is deprecated. Use fill_diagonal instead.", stacklevel=2)
    return fill_diagonal(w, val=val, wsp=wsp)


def fill_diagonal(w, val=1.0, wsp=False):
    """
    Returns a new weights object with values inserted along the main diagonal.

    Parameters
    ----------
    w        : W
               Spatial weights object

    diagonal : float, int or array
               Defines the value(s) to which the weights matrix diagonal should
               be set. If a constant is passed then each element along the
               diagonal will get this value (default is 1.0). An array of length
               w.n can be passed to set explicit values to each element along
               the diagonal (assumed to be in the same order as w.id_order).

    wsp      : boolean
               If True return a thin weights object of the type WSP, if False
               return the standard W object.

    Returns
    -------
    w        : W
               Spatial weights object

    Examples
    --------
    >>> from libpysal.weights import lat2W
    >>> import numpy as np

    Build a basic rook weights matrix, which has zeros on the diagonal, then
    insert ones along the diagonal.

    >>> w = lat2W(5, 5, id_type='string')
    >>> w_const = insert_diagonal(w)
    >>> w['id0'] ==  {'id5': 1.0, 'id1': 1.0}
    True
    >>> w_const['id0'] == {'id5': 1.0, 'id0': 1.0, 'id1': 1.0}
    True

    Insert different values along the main diagonal.

    >>> diag = np.arange(100, 125)
    >>> w_var = insert_diagonal(w, diag)
    >>> w_var['id0'] == {'id5': 1.0, 'id0': 100.0, 'id1': 1.0}
    True
    """

    w_new = copy.deepcopy(w.sparse)
    w_new = w_new.tolil()
    if issubclass(type(val), np.ndarray):
        if w.n != val.shape[0]:
            raise Exception("shape of w and diagonal do not match")
        w_new.setdiag(val)
    elif isinstance(val, numbers.Number):
        w_new.setdiag([val] * w.n)
    else:
        raise Exception("Invalid value passed to diagonal")
    w_out = WSP(w_new, copy.copy(w.id_order))
    if wsp:
        return w_out
    else:
        return WSP2W(w_out)


def remap_ids(w, old2new, id_order=[], **kwargs):  # noqa: B006
    """
    Remaps the IDs in a spatial weights object.

    Parameters
    ----------
    w        : W
               Spatial weights object
    old2new  : dictionary
               Dictionary where the keys are the IDs in w (i.e. "old IDs") and
               the values are the IDs to replace them (i.e. "new IDs")
    id_order : list
               An ordered list of new IDs, which defines the order of observations when
               iterating over W. If not set then the id_order in w will be
               used.
    **kwargs : keyword arguments
               optional arguments for :class:`pysal.weights.W`

    Returns
    -------
    implicit : W
               Spatial weights object with new IDs

    Examples
    --------
    >>> from libpysal.weights import lat2W
    >>> w = lat2W(3,2)
    >>> w.id_order
    [0, 1, 2, 3, 4, 5]
    >>> w.neighbors[0]
    [2, 1]
    >>> old_to_new = {0:'a', 1:'b', 2:'c', 3:'d', 4:'e', 5:'f'}
    >>> w_new = remap_ids(w, old_to_new)
    >>> w_new.id_order
    ['a', 'b', 'c', 'd', 'e', 'f']
    >>> w_new.neighbors['a']
    ['c', 'b']
    """

    if not isinstance(w, W):
        raise Exception("w must be a spatial weights object")
    new_neigh = {}
    new_weights = {}
    for key, value in list(w.neighbors.items()):
        new_values = [old2new[i] for i in value]
        new_key = old2new[key]
        new_neigh[new_key] = new_values
        new_weights[new_key] = copy.copy(w.weights[key])
    if id_order:
        return W(new_neigh, new_weights, id_order, **kwargs)
    else:
        if w.id_order:
            id_order = [old2new[i] for i in w.id_order]
            return W(new_neigh, new_weights, id_order, **kwargs)
        else:
            return W(new_neigh, new_weights, **kwargs)


def get_ids(in_shps, idVariable):
    """
    Gets the IDs from the DBF file that moves with a given shape file or
    a geopandas.GeoDataFrame.

    Parameters
    ----------
    in_shps      : str or geopandas.GeoDataFrame
                   The input geographic data. Either
                   (1) a path to a shapefile including suffix (str); or
                   (2) a geopandas.GeoDataFrame.
    idVariable   : str
                   name of a column in the shapefile's DBF or the
                   geopandas.GeoDataFrame to use for ids.

    Returns
    -------
    ids          : list
                   a list of IDs

    Examples
    --------
    >>> from libpysal.weights.util import get_ids
    >>> import libpysal
    >>> polyids = get_ids(libpysal.examples.get_path("columbus.shp"), "POLYID")
    >>> polyids[:5]
    [1, 2, 3, 4, 5]

    >>> from libpysal.weights.util import get_ids
    >>> import libpysal
    >>> import geopandas as gpd
    >>> gdf = gpd.read_file(libpysal.examples.get_path("columbus.shp"))
    >>> polyids = gdf["POLYID"]
    >>> polyids[:5]
    0    1
    1    2
    2    3
    3    4
    4    5
    Name: POLYID, dtype: int64
    """

    try:
        if isinstance(in_shps, str):
            dbname = os.path.splitext(in_shps)[0] + ".dbf"
            db = FileIO(dbname)
            cols = db.header
            var = db.by_col[idVariable]
            db.close()
        else:
            cols = list(in_shps.columns)
            var = list(in_shps[idVariable])
        return var

    except OSError:
        msg = (
            f'The shapefile "{in_shps}" appears to be missing its DBF file. '
            f' The DBF file "{dbname}" could not be found.'
        )
        raise OSError(msg) from None
    except (AttributeError, KeyError):
        msg = (
            f'The variable "{idVariable}" not found in the DBF/GDF. '
            f"The the following variables are present: {','.join(cols)}."
        )
        raise KeyError(msg) from None


def get_points_array(iterable):
    """
    Gets a data array of x and y coordinates from a given iterable

    Parameters
    ----------
    iterable      : iterable
                    arbitrary collection of shapes that supports iteration

    Returns
    -------
    points        : array
                    (n, 2)
                    a data array of x and y coordinates

    Notes
    -----
    If the given shape file includes polygons,
    this function returns x and y coordinates of the polygons' centroids
    """
    first_choice, backup = tee(iterable)
    try:
        if HAS_SHAPELY:
            data = np.vstack(
                [
                    np.array(shape.centroid.coords)[0]
                    if isinstance(shape, BaseGeometry)
                    else np.array(shape.centroid)
                    for shape in first_choice
                ]
            )
        else:
            data = np.vstack([np.array(shape.centroid) for shape in first_choice])
    except AttributeError:
        data = np.vstack(list(backup))
    return data


def get_points_array_from_shapefile(shapefile):
    """
    Gets a data array of x and y coordinates from a given shapefile.

    Parameters
    ----------
    shapefile     : string
                    name of a shape file including suffix

    Returns
    -------
    points        : array
                    (n, 2)
                    a data array of x and y coordinates

    Notes
    -----
    If the given shape file includes polygons,
    this function returns x and y coordinates of the polygons' centroids

    Examples
    --------
    Point shapefile

    >>> import libpysal
    >>> from libpysal.weights.util import get_points_array_from_shapefile
    >>> xy = get_points_array_from_shapefile(
    ...     libpysal.examples.get_path('juvenile.shp')
    ... )
    >>> xy[:3]
    array([[94., 93.],
           [80., 95.],
           [79., 90.]])

    Polygon shapefile

    >>> xy = get_points_array_from_shapefile(
    ...     libpysal.examples.get_path('columbus.shp')
    ... )
    >>> xy[:3]
    array([[ 8.82721847, 14.36907602],
           [ 8.33265837, 14.03162401],
           [ 9.01226541, 13.81971908]])
    """

    f = FileIO(shapefile)
    data = get_points_array(f)
    return data


def min_threshold_distance(data, p=2):
    """
    Get the maximum nearest neighbor distance.

    Parameters
    ----------
    data    : array
              (n,k) or KDTree where KDtree.data is array (n,k)
              n observations on k attributes
    p       : float
              Minkowski p-norm distance metric parameter:
              1<=p<=infinity
              2: Euclidean distance
              1: Manhattan distance

    Returns
    -------
    nnd    : float
             maximum nearest neighbor distance between the n observations

    Examples
    --------
    >>> from libpysal.weights.util import min_threshold_distance
    >>> import numpy as np
    >>> x, y = np.indices((5, 5))
    >>> x.shape = (25, 1)
    >>> y.shape = (25, 1)
    >>> data = np.hstack([x, y])
    >>> min_threshold_distance(data)
    1.0
    """
    if issubclass(type(data), scipy.spatial.KDTree):
        kd = data
        data = kd.data
    else:
        kd = KDTree(data)
    nn = kd.query(data, k=2, p=p)
    nnd = nn[0].max(axis=0)[1]
    return nnd


def lat2SW(nrows=3, ncols=5, criterion="rook", row_st=False):
    """
    Create a sparse W matrix for a regular lattice.

    Parameters
    ----------
    nrows   : int
              number of rows
    ncols   : int
              number of columns
    rook    : {"rook", "queen", "bishop"}
              type of contiguity. Default is rook.
    row_st  : boolean
              If True, the created sparse W object is row-standardized so
              every row sums up to one. Defaults to False.

    Returns
    -------
    w : scipy.sparse.dia_matrix
        instance of a scipy sparse matrix

    Notes
    -----
    Observations are row ordered: first k observations are in row 0,
    next k in row 1, and so on. This method directly creates the W
    matrix using the strucuture of the contiguity type.

    Examples
    --------
    >>> from libpysal.weights import lat2SW
    >>> w9 = lat2SW(3,3)
    >>> w9[0,1] == 1
    True
    >>> w9[3,6] == 1
    True
    >>> w9r = lat2SW(3,3, row_st=True)
    >>> w9r[3,6] == 1./3
    True
    """

    n = nrows * ncols
    diagonals = []
    offsets = []
    if criterion == "rook" or criterion == "queen":
        d = np.ones((1, n))
        for i in range(ncols - 1, n, ncols):
            d[0, i] = 0
        diagonals.append(d)
        offsets.append(-1)

        d = np.ones((1, n))
        diagonals.append(d)
        offsets.append(-ncols)

    if criterion == "queen" or criterion == "bishop":
        d = np.ones((1, n))
        for i in range(0, n, ncols):
            d[0, i] = 0
        diagonals.append(d)
        offsets.append(-(ncols - 1))

        d = np.ones((1, n))
        for i in range(ncols - 1, n, ncols):
            d[0, i] = 0
        diagonals.append(d)
        offsets.append(-(ncols + 1))
    data = np.concatenate(diagonals)
    offsets = np.array(offsets)
    m = sparse.dia_matrix((data, offsets), shape=(n, n), dtype=np.int8)
    m = m + m.T
    if row_st:
        m = sparse.spdiags(1.0 / m.sum(1).T, 0, *m.shape) * m
    m = m.tocsc()
    m.eliminate_zeros()
    return m


def write_gal(file, k=10):
    with open(file, "w") as f:
        n = k * k
        f.write("0 %d" % n)
        for i in range(n):
            neighs = [i - i, i + 1, i - k, i + k]
            neighs = [j for j in neighs if j >= 0 and j < n]
            f.write("\n%d %d\n" % (i, len(neighs)))
            f.write(" ".join(map(str, neighs)))
        f.close()


def neighbor_equality(w1, w2):
    """
    Test if the neighbor sets are equal between two weights objects

    Parameters
    ----------
    w1 : W
        instance of spatial weights class W
    w2 : W
        instance of spatial weights class W

    Returns
    -------
    bool

    Notes
    -----
    Only set membership is evaluated, no check of the weight values is carried out.

    Examples
    --------
    >>> from libpysal.weights.util import neighbor_equality
    >>> from libpysal.weights import lat2W, W
    >>> w1 = lat2W(3,3)
    >>> w2 = lat2W(3,3)
    >>> neighbor_equality(w1, w2)
    True
    >>> w3 = lat2W(5,5)
    >>> neighbor_equality(w1, w3)
    False
    >>> n4 = w1.neighbors.copy()
    >>> n4[0] = [1]
    >>> n4[1] = [4, 2]
    >>> w4 = W(n4)
    >>> neighbor_equality(w1, w4)
    False
    >>> n5 = w1.neighbors.copy()
    >>> n5[0]
    [3, 1]
    >>> n5[0] = [1, 3]
    >>> w5 = W(n5)
    >>> neighbor_equality(w1, w5)
    True
    """
    n1 = w1.neighbors
    n2 = w2.neighbors
    ids_1 = set(n1.keys())
    ids_2 = set(n2.keys())
    if ids_1 != ids_2:
        return False
    return all(set(w1.neighbors[i]) == set(w2.neighbors[i]) for i in ids_1)


def isKDTree(obj):
    """
    This is a utility function to determine whether or not an object is a
    KDTree, since KDTree and cKDTree have no common parent type
    """
    return any([issubclass(type(obj), KDTYPE) for KDTYPE in KDTREE_TYPES])  # noqa: C419


def attach_islands(w, w_knn1, **kwargs):
    """
    Attach nearest neighbor to islands in spatial weight w.

    Parameters
    ----------
    w            : libpysal.weights.W
                   pysal spatial weight object (unstandardized).
    w_knn1       : libpysal.weights.W
                   Nearest neighbor pysal spatial weight object (k=1).
    **kwargs     : keyword arguments
                   optional arguments for :class:`pysal.weights.W`

    Returns
    -------
                 : libpysal.weights.W
                   pysal spatial weight object w without islands.

    Examples
    --------
    >>> from libpysal.weights import lat2W, Rook, KNN, attach_islands
    >>> import libpysal
    >>> w = Rook.from_shapefile(libpysal.examples.get_path('10740.shp'))
    >>> w.islands
    [163]
    >>> w_knn1 = KNN.from_shapefile(libpysal.examples.get_path('10740.shp'),k=1)
    >>> w_attach = attach_islands(w, w_knn1)
    >>> w_attach.islands
    []
    >>> w_attach[w.islands[0]]
    {166: 1.0}
    """

    neighbors, weights = copy.deepcopy(w.neighbors), copy.deepcopy(w.weights)
    if not len(w.islands):
        print("There are no disconnected observations (no islands)!")
        return w
    else:
        for island in w.islands:
            nb = w_knn1.neighbors[island][0]
            if isinstance(island, float):
                nb = float(nb)
            neighbors[island] = [nb]
            weights[island] = [1.0]
            neighbors[nb] = neighbors[nb] + [island]
            weights[nb] = weights[nb] + [1.0]
        return W(neighbors, weights, id_order=w.id_order, **kwargs)


def nonplanar_neighbors(w, geodataframe, tolerance=0.001, **kwargs):
    """
    Detect neighbors for non-planar polygon collections

    Parameters
    ----------
    w:   pysal W
         A spatial weights object with reported islands
    geodataframe: GeoDataframe
                  The polygon dataframe from which w was constructed.
    tolerance: float
               The percentage of the minimum horizontal or vertical extent (minextent)
               of the dataframe to use in defining  a buffering distance to allow for
               fuzzy contiguity detection. The buffering distance is equal to
               tolerance*minextent.
    **kwargs:  keyword arguments
               optional arguments for :class:`pysal.weights.W`

    Attributes
    ----------
    non_planar_joins : dictionary
               Stores the new joins detected. Key is the id of the focal unit,
               value is a list of neighbor ids.

    Returns
    -------
    w: pysal W
       Spatial weights object that encodes fuzzy neighbors.
       This will have an attribute `non_planar_joins` to indicate
       what new joins were detected.

    Notes
    -----
    This relaxes the notion of contiguity neighbors for the case of shapefiles
    that violate the condition of planar enforcement. It handles three types
    of conditions present in such files that would result in islands when using
    the regular PySAL contiguity methods. The first are edges for nearby
    polygons that should be shared, but are digitized separately for the
    individual polygons and the resulting edges do not coincide, but instead
    the edges intersect. The second case is similar to the first, only the
    resultant edges do not intersect but are "close". The final case arises
    when one polygon is "inside" a second polygon but is not encoded to
    represent a hole in the containing polygon.

    The buffering check assumes the geometry coordinates are projected.

    Examples
    --------
    >>> import geopandas as gpd
    >>> import libpysal
    >>> df = gpd.read_file(libpysal.examples.get_path('map_RS_BR.shp'))
    >>> w = libpysal.weights.Queen.from_dataframe(df)
    >>> w.islands
    [0, 4, 23, 27, 80, 94, 101, 107, 109, 119, 122, 139, 169, 175, 223, 239, 247, 253, 254, 255, 256, 261, 276, 291, 294, 303, 321, 357, 374]
    >>> wnp = libpysal.weights.nonplanar_neighbors(w, df)
    >>> wnp.islands
    []
    >>> w.neighbors[0]
    []
    >>> wnp.neighbors[0]
    [23, 59, 152, 239]
    >>> wnp.neighbors[23]
    [0, 45, 59, 107, 152, 185, 246]

    Also see `nonplanarweights.ipynb`

    References
    ----------
    Planar Enforcement:
    http://ibis.geog.ubc.ca/courses/klink/gis.notes/ncgia/u12.html#SEC12.6
    """  # noqa: E501

    gdf = geodataframe
    assert gdf.sindex, (
        "GeoDataFrame must have a spatial index. "
        "Please make sure you have `libspatialindex` installed"
    )
    islands = w.islands
    joins = copy.deepcopy(w.neighbors)
    candidates = gdf.geometry
    fixes = defaultdict(list)

    # first check for intersecting polygons
    for island in islands:
        focal = gdf.iloc[island].geometry
        neighbors = [
            j
            for j, candidate in enumerate(candidates)
            if focal.intersects(candidate) and j != island
        ]
        if len(neighbors) > 0:
            for neighbor in neighbors:
                if neighbor not in joins[island]:
                    fixes[island].append(neighbor)
                    joins[island].append(neighbor)
                if island not in joins[neighbor]:
                    fixes[neighbor].append(island)
                    joins[neighbor].append(island)

    # if any islands remain, dilate them and check for intersection
    if islands:
        x0, y0, x1, y1 = gdf.total_bounds
        distance = tolerance * min(x1 - x0, y1 - y0)
        for island in islands:
            dilated = gdf.iloc[island].geometry.buffer(distance)
            neighbors = [
                j
                for j, candidate in enumerate(candidates)
                if dilated.intersects(candidate) and j != island
            ]
            if len(neighbors) > 0:
                for neighbor in neighbors:
                    if neighbor not in joins[island]:
                        fixes[island].append(neighbor)
                        joins[island].append(neighbor)
                    if island not in joins[neighbor]:
                        fixes[neighbor].append(island)
                        joins[neighbor].append(island)

    w = W(joins, **kwargs)
    w.non_planar_joins = fixes
    return w


@requires("geopandas")
def fuzzy_contiguity(
    gdf,
    tolerance=0.005,
    buffering=False,
    drop=True,
    buffer=None,
    predicate="intersects",
    **kwargs,
):
    """
    Fuzzy contiguity spatial weights

    Parameters
    ----------
    gdf:   GeoDataFrame
    tolerance: float
               The percentage of the length of the minimum side of the bounding
               rectangle for the GeoDataFrame to use in determining the buffering
               distance.
    buffering: boolean
               If False (default) joins will only be detected for features that
               intersect (touch, contain, within). If True then features will be
               buffered and intersections will be based on buffered features.
    drop: boolean
          If True (default), the buffered features are removed from the GeoDataFrame.
          If False, buffered features are added to the GeoDataFrame.
    buffer : float
             Specify exact buffering distance. Ignores `tolerance`.
    predicate : {'intersects', 'within', 'contains', 'overlaps', 'crosses', 'touches'}
                The predicate to use for determination of neighbors. Default is
                'intersects'. If None is passed, neighbours are determined based on
                the intersection of bounding boxes.
    **kwargs: keyword arguments
              optional arguments for :class:`pysal.weights.W`

    Returns
    -------
    w:  PySAL W
        Spatial weights based on fuzzy contiguity. Weights are binary.

    Examples
    --------
    >>> import libpysal
    >>> from libpysal.weights import fuzzy_contiguity
    >>> import geopandas as gpd
    >>> rs = libpysal.examples.get_path('map_RS_BR.shp')
    >>> rs_df = gpd.read_file(rs)
    >>> wq = libpysal.weights.Queen.from_dataframe(rs_df)
    >>> len(wq.islands)
    29
    >>> wq[0]
    {}
    >>> wf = fuzzy_contiguity(rs_df)
    >>> wf.islands
    []
    >>> wf[0] == dict({239: 1.0, 59: 1.0, 152: 1.0, 23: 1.0, 107: 1.0})
    True

    Example needing to use buffering

    >>> from shapely.geometry import Polygon
    >>> p0 = Polygon([(0,0), (10,0), (10,10)])
    >>> p1 = Polygon([(10,1), (10,2), (15,2)])
    >>> p2 = Polygon([(12,2.001), (14, 2.001), (13,10)])
    >>> gs = gpd.GeoSeries([p0,p1,p2])
    >>> gdf = gpd.GeoDataFrame(geometry=gs)
    >>> wf = fuzzy_contiguity(gdf)
    >>> wf.islands
    [2]
    >>> wfb = fuzzy_contiguity(gdf, buffering=True)
    >>> wfb.islands
    []
    >>> wfb[2]
    {1: 1.0}

    Example with a custom index

    >>> rs_df_ix = rs_df.set_index("NM_MUNICIP")
    >>> wf_ix = fuzzy_contiguity(rs_df)
    >>> wf_ix.neighbors["TAVARES"]
    ['SÃO JOSÉ DO NORTE', 'MOSTARDAS']

    Notes
    -----
    This relaxes the notion of contiguity neighbors for the case of feature
    collections that violate the condition of planar enforcement. It handles
    three types of conditions present in such collections that would result in
    islands when using the regular PySAL contiguity methods. The first are
    edges for nearby polygons that should be shared, but are digitized
    separately for the individual polygons and the resulting edges do not
    coincide, but instead the edges intersect. The second case is similar to
    the first, only the resultant edges do not intersect but are "close". The
    final case arises when one polygon is "inside" a second polygon but is not
    encoded to represent a hole in the containing polygon.

    Detection of the second case will require setting buffering=True
    and exploring different values for tolerance.

    The buffering check assumes the geometry coordinates are projected.

    References
    ----------
    Planar Enforcement:
    http://ibis.geog.ubc.ca/courses/klink/gis.notes/ncgia/u12.html#SEC12.6
    """
    if buffering:
        if not buffer:
            # buffer each shape
            minx, miny, maxx, maxy = gdf.total_bounds
            buffer = tolerance * 0.5 * abs(min(maxx - minx, maxy - miny))
        # create new geometry column
        new_geometry = gdf.geometry.buffer(buffer)
        gdf["_buffer"] = new_geometry
        old_geometry_name = gdf.geometry.name
        gdf.set_geometry("_buffer", inplace=True)

    neighbors = {}
    if GPD_013:
        # query tree based on set predicate
        inp, res = gdf.sindex.query(gdf.geometry, predicate=predicate)
    else:
        inp, res = gdf.sindex.query_bulk(gdf.geometry, predicate=predicate)
    # remove self hits
    itself = inp == res
    inp = inp[~itself]
    res = res[~itself]

    # extract index values of neighbors
    for i, ix in enumerate(gdf.index):
        ids = gdf.index[res[inp == i]].tolist()
        neighbors[ix] = ids

    if buffering:
        gdf.set_geometry(old_geometry_name, inplace=True)
        if drop:
            gdf.drop(columns=["_buffer"], inplace=True)

    return W(neighbors, **kwargs)


if __name__ == "__main__":
    assert (lat2W(5, 5).sparse.todense() == lat2SW(5, 5).todense()).all()
    assert (lat2W(5, 3).sparse.todense() == lat2SW(5, 3).todense()).all()
    assert (
        lat2W(5, 3, rook=False).sparse.todense() == lat2SW(5, 3, "queen").todense()
    ).all()
    assert (
        lat2W(50, 50, rook=False).sparse.todense() == lat2SW(50, 50, "queen").todense()
    ).all()



import geodatasets
import geopandas
import numpy as np
import pytest

from libpysal import graph


@pytest.mark.network
class TestSummary:
    def setup_method(self):
        self.nybb = geopandas.read_file(geodatasets.get_path("nybb"))
        self.G = graph.Graph.build_contiguity(self.nybb)
        self.summary = self.G.summary(True)
        self.no_asymmetries = self.G.summary()

    def test_exactness(self):
        assert self.summary.n_nodes == 5
        assert self.summary.n_edges == 10
        assert self.summary.n_components == 2
        assert self.summary.n_isolates == 1
        assert self.summary.nonzero == 10
        assert self.summary.pct_nonzero == 44.0
        assert self.summary.n_asymmetries == 0
        assert self.summary.cardinalities_mean == 2.0
        assert self.summary.cardinalities_std == 1.224744871391589
        assert self.summary.cardinalities_min == 0
        assert self.summary.cardinalities_25 == 2
        assert self.summary.cardinalities_50 == 2
        assert self.summary.cardinalities_75 == 3
        assert self.summary.cardinalities_max == 3
        assert self.summary.weights_mean == 1
        assert self.summary.weights_std == 0
        assert self.summary.weights_min == 1
        assert self.summary.weights_25 == 1
        assert self.summary.weights_50 == 1
        assert self.summary.weights_75 == 1
        assert self.summary.weights_max == 1
        assert self.summary.s0 == 10
        assert self.summary.s1 == 20.0
        assert self.summary.s2 == 104.0
        np.testing.assert_array_equal(np.array([0, 3, 2, 3, 2]), self.summary.diag_g2)
        np.testing.assert_array_equal(np.array([0, 3, 2, 3, 2]), self.summary.diag_gtg)
        np.testing.assert_array_equal(
            np.array([0, 6, 4, 6, 4]), self.summary.diag_gtg_gg
        )
        assert self.summary.trace_g2 == 10.0
        assert self.summary.trace_gtg == 10.0
        assert self.summary.trace_gtg_gg == 20.0

    def test_repr(self):
        expected = """Graph Summary Statistics
========================
Graph indexed by:
 [0, 1, 2, 3, 4]
==============================================================
Number of nodes:                                             5
Number of edges:                                            10
Number of connected components:                              2
Number of isolates:                                          1
Number of non-zero edges:                                   10
Percentage of non-zero edges:                           44.00%
Number of asymmetries:                                       0
--------------------------------------------------------------
Cardinalities
==============================================================
Mean:                       2    25%:                        2
Standard deviation:         1    50%:                        2
Min:                        0    75%:                        3
Max:                        3
--------------------------------------------------------------
Weights
==============================================================
Mean:                       1    25%:                        1
Standard deviation:         0    50%:                        1
Min:                        1    75%:                        1
Max:                        1
--------------------------------------------------------------
Sum of weights
==============================================================
S0:                                                         10
S1:                                                         20
S2:                                                        104
--------------------------------------------------------------
Traces
==============================================================
GG:                                                         10
G'G:                                                        10
G'G + GG:                                                   20
"""
        assert self.summary.__repr__() == expected

    def test_html_repr(self):
        expected = """
            <table>
                <caption>Graph Summary Statistics</caption>
                <tr>
                    <td>Number of nodes:</td>
                    <td>           5</td>
                </tr>
                <tr>
                    <td>Number of edges:</td>
                    <td>          10</td>
                </tr>
                <tr>
                    <td>Number of connected components:</td>
                    <td>           2</td>
                </tr>
                <tr>
                    <td>Number of isolates:</td>
                    <td>           1</td>
                </tr>
                <tr>
                    <td>Number of non-zero edges:</td>
                    <td>          10</td>
                </tr>
                <tr>
                    <td>Percentage of non-zero edges:</td>
                    <td>      44.00%</td>
                </tr>
                <tr>
                    <td>Number of asymmetries:</td>
                    <td>           0</td>
                </tr>
            </table>
            <table>
                <caption>Cardinalities</caption>
                <tr>
                    <td>Mean:</td>
                    <td>        2</td>
                    <td>25%:</td>
                    <td>        2</td>
                </tr>
                <tr>
                    <td>Standard deviation:</td>
                    <td>        1</td>
                    <td>50%</td>
                    <td>        2
                    <td>
                </tr>
                <tr>
                    <td>Min:</td>
                    <td>        0</td>
                    <td>75%:</td>
                    <td>        3</td>
                </tr>
                <tr>
                    <td>Max:</td>
                    <td>        3</td>
                </tr>
            </table>
            <table>
                <caption>Weights</caption>
                <tr>
                    <td>Mean:</td>
                    <td>        1</td>
                    <td>25%:</td>
                    <td>        1</td>
                </tr>
                <tr>
                    <td>Standard deviation:</td>
                    <td>        0</td>
                    <td>50%</td>
                    <td>        1
                <tr>
                    <td>Min:</td>
                    <td>        1</td>
                    <td>75%:</td>
                    <td>        1</td>
                </tr>
                <tr>
                    <td>Max:</td>
                    <td>        1</td>
                </tr>
            </table>
            <table>
                <caption>Sum of weights and Traces</caption>
                <tr>
                    <td>S0:</td>
                    <td>          10</td>
                    <td>GG:</td>
                    <td>          10</td>
                </tr>
                <tr>
                    <td>S1:</td>
                    <td>          20</td>
                    <td>G'G:</td>
                    <td>          10</td>
                </tr>
                <tr>
                    <td>S3:</td>
                    <td>         104</td>
                    <td>G'G + GG:</td>
                    <td>          20</td>
                </tr>
            </table>
            <div>
                Graph indexed by: <code>[0, 1, 2, 3, 4]</code>
            </div>
            """

        assert self.summary._repr_html_() == expected

    def test_no_asymmetries(self):
        assert not hasattr(self.no_asymmetries, "n_asymmetries")
        _ = self.no_asymmetries.__repr__()
        _ = self.no_asymmetries._repr_html_()



import os
import tempfile

import pytest

from .... import examples as pysal_examples
from ...fileio import FileIO
from ..geobugs_txt import GeoBUGSTextIO


class TesttestGeoBUGSTextIO:
    def setup_method(self):
        self.test_file_scot = test_file_scot = pysal_examples.get_path("geobugs_scot")
        self.test_file_col = test_file_col = pysal_examples.get_path(
            "spdep_listw2WB_columbus"
        )
        self.obj_scot = GeoBUGSTextIO(test_file_scot, "r")
        self.obj_col = GeoBUGSTextIO(test_file_col, "r")

    def test_close(self):
        for obj in [self.obj_scot, self.obj_col]:
            f = obj
            f.close()
            pytest.raises(ValueError, f.read)

    def test_read(self):
        w_scot = self.obj_scot.read()
        assert w_scot.n == 56
        assert w_scot.mean_neighbors == 4.1785714285714288
        assert list(w_scot[1].values()) == [1.0, 1.0, 1.0]

        w_col = self.obj_col.read()
        assert w_col.n == 49
        assert w_col.mean_neighbors == 4.6938775510204085
        assert list(w_col[1].values()) == [0.5, 0.5]

    def test_seek(self):
        self.test_read()
        pytest.raises(StopIteration, self.obj_scot.read)
        pytest.raises(StopIteration, self.obj_col.read)
        self.obj_scot.seek(0)
        self.obj_col.seek(0)
        self.test_read()

    def test_write(self):
        for obj in [self.obj_scot, self.obj_col]:
            w = obj.read()
            f = tempfile.NamedTemporaryFile(suffix="")
            fname = f.name
            f.close()
            o = FileIO(fname, "w", "geobugs_text")
            o.write(w)
            o.close()
            wnew = FileIO(fname, "r", "geobugs_text").read()
            assert wnew.pct_nonzero == w.pct_nonzero
            os.remove(fname)



import numpy as np


class GraphSummary:
    r"""Graph Summary

    An object containing the statistical attributes summarising the Graph and its basic
    properties.

    Attributes
    ----------
    n_nodes : int
        number of Graph nodes
    n_edges : int
        number of Graph edges
    n_components : int
        number of connected components
    n_isolates : int
        number of isolates (nodes with no neighbors)
    nonzero : int
        number of edges with nonzero weight
    pct_nonzero : float
        percentage of nonzero weights
    n_asymmetries : int
        number of intrinsic asymmetries
    cardinalities_mean : float
        mean number of neighbors
    cardinalities_std : float
        standard deviation of number of neighbors
    cardinalities_min : float
        minimal number of neighbors
    cardinalities_25 : float
        25th percentile of number of neighbors
    cardinalities_50 : float
        50th percentile (median) of number of neighbors
    cardinalities_75 : float
        75th percentile of number of neighbors
    cardinalities_max : float
        maximal number of neighbors
    weights_mean : float
        mean edge weight
    weights_std : float
        standard deviation of  edge weights
    weights_min : float
        minimal edge weight
    weights_25 : float
        25th percentile of edge weights
    weights_50 : float
        50th percentile (median) of edge weights
    weights_75 : float
        75th percentile of edge weights
    weights_max : float
        maximal edge weight
    s0 : float
        S0 (global) sum of weights

        ``s0`` is defined as

        .. math::
               s0=\sum_i \sum_j w_{i,j}

        :attr:`s0`, :attr:`s1`, and :attr:`s2` reflect interaction between observations
        and are used to compute standard errors for spatial autocorrelation estimators.


    s1 : float
        S1 sum of weights

        ``s1`` is defined as

        .. math::
               s1=1/2 \sum_i \sum_j \Big(w_{i,j} + w_{j,i}\Big)^2

        :attr:`s0`, :attr:`s1`, and :attr:`s2` reflect interaction between observations
        and are used to compute standard errors for spatial autocorrelation estimators.

    s2 : float
        S2 sum of weights

        ``s2`` is defined as

        .. math::
                s2=\sum_j \Big(\sum_i w_{i,j} + \sum_i w_{j,i}\Big)^2

        :attr:`s0`, :attr:`s1`, and :attr:`s2` reflect interaction between observations
        and are used to compute standard errors for spatial autocorrelation estimators.

    diag_g2 : np.ndarray
        diagonal of :math:`GG`
    diag_gtg : np.ndarrray
        diagonal of :math:`G^{'}G`
    diag_gtg_gg : np.ndarray
        diagonal of :math:`G^{'}G + GG`
    trace_g2 : np.ndarray
        trace of :math:`GG`
    trace_gtg : np.ndarrray
        trace of :math:`G^{'}G`
    trace_gtg_gg : np.ndarray
        trace of :math:`G^{'}G + GG`

    Examples
    --------
    >>> import geopandas as gpd
    >>> from geodatasets import get_path
    >>> nybb = gpd.read_file(get_path("nybb")).set_index("BoroName")
    >>> nybb
                    BoroCode  ...                                           geometry
    BoroName                 ...
    Staten Island         5  ...  MULTIPOLYGON (((970217.022 145643.332, 970227....
    Queens                4  ...  MULTIPOLYGON (((1029606.077 156073.814, 102957...
    Brooklyn              3  ...  MULTIPOLYGON (((1021176.479 151374.797, 102100...
    Manhattan             1  ...  MULTIPOLYGON (((981219.056 188655.316, 980940....
    Bronx                 2  ...  MULTIPOLYGON (((1012821.806 229228.265, 101278...
    [5 rows x 4 columns]

    >>> contiguity = graph.Graph.build_contiguity(nybb)
    >>> contiguity
    <Graph of 5 nodes and 10 nonzero edges indexed by
        ['Staten Island', 'Queens', 'Brooklyn', 'Manhattan', 'Bronx']>

    >>> summary = contiguity.summary(asymmetries=True)
    >>> summary
    Graph Summary Statistics
    ========================
    Graph indexed by:
    ['Staten Island', 'Queens', 'Brooklyn', 'Manhattan', 'Bronx']
    ==============================================================
    number of nodes:                                             5
    number of edges:                                            10
    number of connected components:                              2
    number of isolates:                                          1
    number of non-zero edges:                                   10
    Percentage of non-zero edges:                           44.00%
    number of asymmetries:                                       0
    --------------------------------------------------------------
    Cardinalities
    ==============================================================
    Mean:                       2    25%:                        2
    Standard deviation:         1    50%:                        2
    Min:                        0    75%:                        3
    Max:                        3
    --------------------------------------------------------------
    Weights
    ==============================================================
    Mean:                       1    25%:                        1
    Standard deviation:         0    50%:                        1
    Min:                        1    75%:                        1
    Max:                        1
    --------------------------------------------------------------
    Sum of weights
    ==============================================================
    S0:                                                         10
    S1:                                                         20
    S2:                                                        104
    --------------------------------------------------------------
    Traces
    ==============================================================
    GG:                                                         10
    G'G:                                                        10
    G'G + GG:                                                   20

    >>> summary.s1
    20
    """

    def __init__(self, graph, asymmetries=False):
        """Create GraphSummary

        Parameters
        ----------
        graph : Graph
        asymmetries : bool
            whether to compute ``n_asymmetries``, which is considerably more expensive
            than the other attributes. By default False.
        """
        self._graph = graph
        self.asymmetries = asymmetries

        self.n_nodes = self._graph.n_nodes  # number of nodes / observations
        self.n_edges = self._graph.n_edges  # number of edges excluding isolates
        self.n_components = self._graph.n_components
        self.n_isolates = len(self._graph.isolates)

        # nonzero
        self.nonzero = self._graph.nonzero
        self.pct_nonzero = self._graph.pct_nonzero

        # intrinsic assymetries
        if asymmetries:
            self.n_asymmetries = len(self._graph.asymmetry())

        # statistics of cardinalities
        card_stats = self._graph.cardinalities.describe()
        self.cardinalities_mean = card_stats["mean"]
        self.cardinalities_std = card_stats["std"]
        self.cardinalities_min = card_stats["min"]
        self.cardinalities_25 = card_stats["25%"]
        self.cardinalities_50 = card_stats["50%"]
        self.cardinalities_75 = card_stats["75%"]
        self.cardinalities_max = card_stats["max"]

        # statistics of weights
        weights_stats = self._graph._adjacency.drop(self._graph.isolates).describe()
        self.weights_mean = weights_stats["mean"]
        self.weights_std = weights_stats["std"]
        self.weights_min = weights_stats["min"]
        self.weights_25 = weights_stats["25%"]
        self.weights_50 = weights_stats["50%"]
        self.weights_75 = weights_stats["75%"]
        self.weights_max = weights_stats["max"]

        # sum of weights
        self.s0 = self._s0()
        self.s1 = self._s1()
        self.s2 = self._s2()

        # diags
        self.diag_g2 = self._diag_g2()
        self.diag_gtg = self._diag_gtg()
        self.diag_gtg_gg = self._diag_gtg_gg()

        # traces
        self.trace_g2 = self.diag_g2.sum()
        self.trace_gtg = self.diag_gtg.sum()
        self.trace_gtg_gg = self.diag_gtg_gg.sum()

    def __repr__(self):
        n_asymmetries = f"{self.n_asymmetries:>12.0f}" if self.asymmetries else "NA"
        return f"""Graph Summary Statistics
{'='*24}
Graph indexed by:
 {self._graph._get_ids_repr(57)}
{'='*62}
{"Number of nodes:":<50}{self.n_nodes:>12.0f}
{"Number of edges:":<50}{self.n_edges:>12.0f}
{"Number of connected components:":<50}{self.n_components:>12.0f}
{"Number of isolates:":<50}{self.n_isolates:12.0f}
{"Number of non-zero edges:":<50}{self.nonzero:>12.0f}
{"Percentage of non-zero edges:":<50}{self.pct_nonzero:>11.2f}%
{"Number of asymmetries:":<50}{n_asymmetries}
{'-'*62}
Cardinalities
{'='*62}
{"Mean:":<20}{self.cardinalities_mean:>9.0f}    {"25%:":<20}{self.cardinalities_25:>9.0f}
{"Standard deviation:":<20}{self.cardinalities_std:>9.0f}    {"50%:":<20}{self.cardinalities_50:>9.0f}
{"Min:":<20}{self.cardinalities_min:>9.0f}    {"75%:":<20}{self.cardinalities_75:>9.0f}
{"Max:":<20}{self.cardinalities_max:>9.0f}
{'-'*62}
Weights
{'='*62}
{"Mean:":<20}{self.weights_mean:>9.0f}    {"25%:":<20}{self.weights_25:>9.0f}
{"Standard deviation:":<20}{self.weights_std:>9.0f}    {"50%:":<20}{self.weights_50:>9.0f}
{"Min:":<20}{self.weights_min:>9.0f}    {"75%:":<20}{self.weights_75:>9.0f}
{"Max:":<20}{self.weights_max:>9.0f}
{'-'*62}
Sum of weights
{'='*62}
{"S0:":<50}{self.s0:>12.0f}
{"S1:":<50}{self.s1:>12.0f}
{"S2:":<50}{self.s2:>12.0f}
{'-'*62}
Traces
{'='*62}
{"GG:":<50}{self.trace_g2:>12.0f}
{"G'G:":<50}{self.trace_gtg:>12.0f}
{"G'G + GG:":<50}{self.trace_gtg_gg:>12.0f}
"""  # noqa: E501

    def _repr_html_(self):
        n_asymmetries = f"{self.n_asymmetries:12.0f}" if self.asymmetries else "NA"
        return f"""
            <table>
                <caption>Graph Summary Statistics</caption>
                <tr>
                    <td>Number of nodes:</td>
                    <td>{self.n_nodes:12.0f}</td>
                </tr>
                <tr>
                    <td>Number of edges:</td>
                    <td>{self.n_edges:12.0f}</td>
                </tr>
                <tr>
                    <td>Number of connected components:</td>
                    <td>{self.n_components:12.0f}</td>
                </tr>
                <tr>
                    <td>Number of isolates:</td>
                    <td>{self.n_isolates:12.0f}</td>
                </tr>
                <tr>
                    <td>Number of non-zero edges:</td>
                    <td>{self.nonzero:12.0f}</td>
                </tr>
                <tr>
                    <td>Percentage of non-zero edges:</td>
                    <td>{self.pct_nonzero:11.2f}%</td>
                </tr>
                <tr>
                    <td>Number of asymmetries:</td>
                    <td>{n_asymmetries}</td>
                </tr>
            </table>
            <table>
                <caption>Cardinalities</caption>
                <tr>
                    <td>Mean:</td>
                    <td>{self.cardinalities_mean:9.0f}</td>
                    <td>25%:</td>
                    <td>{self.cardinalities_25:9.0f}</td>
                </tr>
                <tr>
                    <td>Standard deviation:</td>
                    <td>{self.cardinalities_std:9.0f}</td>
                    <td>50%</td>
                    <td>{self.cardinalities_50:9.0f}
                    <td>
                </tr>
                <tr>
                    <td>Min:</td>
                    <td>{self.cardinalities_min:9.0f}</td>
                    <td>75%:</td>
                    <td>{self.cardinalities_75:9.0f}</td>
                </tr>
                <tr>
                    <td>Max:</td>
                    <td>{self.cardinalities_max:9.0f}</td>
                </tr>
            </table>
            <table>
                <caption>Weights</caption>
                <tr>
                    <td>Mean:</td>
                    <td>{self.weights_mean:9.0f}</td>
                    <td>25%:</td>
                    <td>{self.weights_25:9.0f}</td>
                </tr>
                <tr>
                    <td>Standard deviation:</td>
                    <td>{self.weights_std:9.0f}</td>
                    <td>50%</td>
                    <td>{self.weights_50:9.0f}
                <tr>
                    <td>Min:</td>
                    <td>{self.weights_min:9.0f}</td>
                    <td>75%:</td>
                    <td>{self.weights_75:9.0f}</td>
                </tr>
                <tr>
                    <td>Max:</td>
                    <td>{self.weights_max:9.0f}</td>
                </tr>
            </table>
            <table>
                <caption>Sum of weights and Traces</caption>
                <tr>
                    <td>S0:</td>
                    <td>{self.s0:12.0f}</td>
                    <td>GG:</td>
                    <td>{self.trace_g2:12.0f}</td>
                </tr>
                <tr>
                    <td>S1:</td>
                    <td>{self.s1:12.0f}</td>
                    <td>G'G:</td>
                    <td>{self.trace_gtg:12.0f}</td>
                </tr>
                <tr>
                    <td>S3:</td>
                    <td>{self.s2:12.0f}</td>
                    <td>G'G + GG:</td>
                    <td>{self.trace_gtg_gg:12.0f}</td>
                </tr>
            </table>
            <div>
                Graph indexed by: <code>{self._graph._get_ids_repr(57)}</code>
            </div>
            """

    def _s0(self):
        r"""helper to get S0 in downstream

         ``s0`` is defined as

        .. math::

               s0=\sum_i \sum_j w_{i,j}

        :attr:`s0`, :attr:`s1`, and :attr:`s2` reflect interaction between observations
        and are used to compute standard errors for spatial autocorrelation estimators.

        Returns
        -------
        float
            global sum of weights
        """
        return self._graph._adjacency.sum()

    def _s1(self):
        r"""S1 sum of weights

        ``s1`` is defined as

        .. math::

               s1=1/2 \sum_i \sum_j \Big(w_{i,j} + w_{j,i}\Big)^2

        :attr:`s0`, :attr:`s1`, and :attr:`s2` reflect interaction between observations
        and are used to compute standard errors for spatial autocorrelation estimators.

        Returns
        -------
        float
            s1 sum of weights
        """
        t = self._graph.sparse.transpose()
        t = t + self._graph.sparse
        t2 = t * t
        return t2.sum() / 2.0

    def _s2(self):
        r"""S2 sum of weights

        ``s2`` is defined as

        .. math::

                s2=\sum_j \Big(\sum_i w_{i,j} + \sum_i w_{j,i}\Big)^2

        :attr:`s0`, :attr:`s1`, and :attr:`s2` reflect interaction between observations
        and are used to compute standard errors for spatial autocorrelation estimators.

        Returns
        -------
        float
            s2 sum of weights
        """
        s = self._graph.sparse
        return (np.array(s.sum(1) + s.sum(0).transpose()) ** 2).sum()

    def _diag_g2(self):
        """Diagonal of :math:`GG`.

        Returns
        -------
        np.ndarray
        """
        return (self._graph.sparse @ self._graph.sparse).diagonal()

    def _diag_gtg(self):
        """Diagonal of :math:`G^{'}G`.

        Returns
        -------
        np.ndarray
        """
        return (self._graph.sparse.transpose() @ self._graph.sparse).diagonal()

    def _diag_gtg_gg(self):
        """Diagonal of :math:`G^{'}G + GG`.

        Returns
        -------
        np.ndarray
        """
        gt = self._graph.sparse.transpose()
        g = self._graph.sparse
        return (gt @ g + g @ g).diagonal()



"""Handle local builtin datasets."""

import os

from .base import get_list_of_files

dirs = [
    "10740",
    "arcgis",
    "baltim",
    "berlin",
    "book",
    "burkitt",
    "calemp",
    "chicago",
    "columbus",
    "desmith",
    "geodanet",
    "georgia",
    "juvenile",
    "Line",
    "mexico",
    "networks",
    "Point",
    "Polygon",
    "Polygon_Holes",
    "sids2",
    "snow_maps",
    "stl",
    "street_net_pts",
    "tests",
    "tokyo",
    "us_income",
    "virginia",
    "wmat",
]


class LocalExample:
    """Builtin pysal example dataset.

    Attributes
    ----------
    name : str
        Example name.
    dirname : str
        Path holding example files
    installed : bool
        If True, example is installed locally, if false it is remote.
    description : str
        Summary of the properties of the example.
    """

    def __init__(self, name, dirname):
        """Initialize LocalExample with name and dirname.

        Parameters
        ---------
        name : str
            Example name.
        dirname: str
             Path to directory holding example files.
        """
        self.name = name
        self.dirname = dirname
        self.installed = True
        self.description = self.get_description()

    def get_file_list(self) -> list:
        """Return a list of file names."""
        return get_list_of_files(self.dirname)

    def get_path(self, file_name: str, verbose=True) -> str:
        """Get path for local file."""
        file_list = self.get_file_list()
        for file_path in file_list:
            base_name = os.path.basename(file_path)
            if file_name == base_name:
                return file_path
        if verbose:
            print(f"{file_name} is not a file in this example")
        return None

    def explain(self):
        """Provide a printed description of the example."""
        description = [f for f in self.get_file_list() if "README.md" in f][0]
        with open(description, encoding="utf8") as f:
            print(f.read())

    def get_description(self) -> str:
        """Dataset description."""
        description = [f for f in self.get_file_list() if "README.md" in f][0]
        with open(description, encoding="utf8") as f:
            lines = f.readlines()
        return lines[3].strip()


builtin_root = os.path.dirname(__file__)

paths = [os.path.join(builtin_root, local) for local in dirs]
paths = zip(dirs, paths, strict=True)

datasets = {}

for name, pth in paths:
    files = get_list_of_files(pth)
    file_names = [os.path.basename(file) for file in files]
    if "README.md" in file_names:
        example = LocalExample(name, pth)
        datasets[name] = example
